from __future__ import absolute_import, division, print_function

import cobra
import cobra.test
from cobra.medium import minimal_medium

from cobra import Model, Reaction, Metabolite, Configuration
from vivarium.library.units import units
from vivarium_cell.processes.derive_globals import AVOGADRO
from vivarium_cell.data.synonyms import get_synonym


EXTERNAL_PREFIX = 'EX_'
DEFAULT_UPPER_BOUND = 100.0

def build_model(stoichiometry, reversible, objective, external_molecules, default_upper_bound=1000):
    model = Model('fba')
    model.compartments = {'c': 'cytoplasm'}

    metabolite_keys = {}
    for reaction_key, chemistry in stoichiometry.items():
        metabolite_keys.update(chemistry)

    metabolites = {
        metabolite: Metabolite(metabolite, name=metabolite, compartment='c')
        for metabolite in list(metabolite_keys.keys())}

    model.add_metabolites(metabolites.values())

    # make reactions
    reactions = {}
    for reaction_key, chemistry in stoichiometry.items():
        reaction = Reaction(reaction_key, name=reaction_key)

        # set reaction bounds
        reaction.upper_bound = default_upper_bound
        if reaction_key in reversible:
            reaction.lower_bound = -reaction.upper_bound

        # make stoichiometry
        reaction_model = {
            metabolites[metabolite]: value
            for metabolite, value in chemistry.items()}
        reaction.add_metabolites(reaction_model)

        reactions[reaction_key] = reaction

    # make exchange reactions for all external_molecules
    for external in external_molecules:
        external_key = EXTERNAL_PREFIX + external
        reaction = Reaction(external_key, name=external_key)

        # set reaction bounds
        reaction.upper_bound = default_upper_bound
        reaction.lower_bound = -default_upper_bound  # TODO -- should exchanges have symmetric bounds by default?

        # make stoichiometry
        reaction_model = {metabolites[external]: -1}
        reaction.add_metabolites(reaction_model)

        reactions[external_key] = reaction

    model.add_reactions(reactions.values())

    model.objective = {
        reactions[reaction_key]: value
        for reaction_key, value in objective.items()}

    return model

def extract_model(model):
    """
    TODO -- where do demands and sinks go?
    demands = model.demands
    sinks = model.sinks

    # boundary reactions include exchanges, demands, sinks
    boundary = model.boundary
    boundary_reactions = [reaction.id for reaction in boundary]
    """

    reactions = model.reactions
    metabolites = model.metabolites
    exchanges = model.exchanges
    objective_expression = model.objective.expression.args

    # get stoichiometry and flux bounds
    stoichiometry = {}
    flux_bounds = {}
    reversible = []
    for reaction in reactions:
        reaction_id = reaction.id
        reaction_metabolites = reaction.metabolites
        bounds = list(reaction.bounds)
        stoichiometry[reaction.id] = {
            metabolite.id: coeff for metabolite, coeff in reaction_metabolites.items()}
        flux_bounds[reaction_id] = bounds
        if not any(b == 0.0 for b in bounds):
            reversible.append(reaction_id)

    # get external molecules and exchange bounds from exchanges
    external_molecules = []
    exchange_bounds = {}
    for reaction in exchanges:
        reaction_metabolites = list(reaction.metabolites.keys())
        assert len(reaction_metabolites) == 1  # only 1 molecule in the exchange reaction
        metabolite_id = reaction_metabolites[0].id
        external_molecules.append(metabolite_id)
        bounds = list(reaction.bounds)
        exchange_bounds[metabolite_id] = bounds

    # get molecular weights
    molecular_weights = {}
    nonexisting_elements = ['R', 'X']
    for metabolite in metabolites:
        if any([e in nonexisting_elements for e in metabolite.elements]):
            continue
        molecular_weights[metabolite.id] = metabolite.formula_weight

    # get objective
    objective = {}
    for expression in objective_expression:
        exp_str = str(expression)
        coeff, reaction_id = exp_str.split('*')
        try:
            reactions.get_by_id(reaction_id)
            objective[reaction_id] = float(coeff)
        except:
            pass

    # get flux scaling factor based on the objective's predicted added mass
    # this adjusts the BiGG FBA bounds to approximate single-cell rates
    target_added_mass = 4.9e-7  # fit to approximate a doubling time of 2520 sec (42 min) in iAF1260b

    solution = model.optimize()
    objective_value = solution.objective_value
    added_mass = 0
    for reaction_id, coeff1 in objective.items():
        for mol_id, coeff2 in stoichiometry[reaction_id].items():
            if coeff2 < 0: # molecule is used to make biomass (negative coefficient)
                mw = molecular_weights.get(mol_id) * (units.g / units.mol)
                count = int(-coeff1 * coeff2 * objective_value)
                mol = count / AVOGADRO
                mol_added_mass = mw * mol
                added_mass += mol_added_mass.to('fg').magnitude

    flux_scaling = target_added_mass / added_mass

    return {
        'stoichiometry': stoichiometry,
        'reversible': reversible,
        'external_molecules': external_molecules,
        'objective': objective,
        'flux_bounds': flux_bounds,
        'exchange_bounds': exchange_bounds,
        'molecular_weights': molecular_weights,
        'flux_scaling': flux_scaling}

def swap_synonyms(model):
    metabolites = model.metabolites

    # swap metabolite ids
    for mol in metabolites:
        mol_id = mol.id
        mol_id = get_synonym(mol_id)
        mol.id = mol_id

class CobraFBA(object):
    """
    This class provides an interface to cobra FBA.
    It can load in BiGG models (http://bigg.ucsd.edu/models) if provided a model_path to a saved JSON BiGG model,
    or load in a novel model specified by stoichiometry, reversibility, and objective.

    TODO (Eran) -- MOMA option is provided, but has not yet been tested.
    """

    cobra_configuration = Configuration()

    def __init__(self, config={}):
        model_path = config.get('model_path')

        # get tolerances
        self.default_tolerance = config.get('default_tolerance', [0.95, 1])
        self.tolerance = config.get('tolerance', {})

        # set MOMA (minimization of metabolic adjustment)
        self.moma = config.get('moma', False)

        if model_path:
            # load a BiGG model
            self.model = cobra.io.load_json_model(model_path)
            swap_synonyms(self.model)
            extract = extract_model(self.model)

            self.stoichiometry = extract['stoichiometry']
            self.reversible = extract['reversible']
            self.external_molecules = extract['external_molecules']
            self.objective = extract['objective']
            self.flux_bounds = extract['flux_bounds']
            self.molecular_weights = extract['molecular_weights']
            self.exchange_bounds = extract['exchange_bounds']
            self.default_upper_bound = DEFAULT_UPPER_BOUND  # TODO -- can this be extracted from model?
            self.flux_scaling = extract['flux_scaling']

        else:
            # create an FBA model from config
            self.stoichiometry = config['stoichiometry']
            self.reversible = config.get('reversible', [])
            self.external_molecules = config['external_molecules']
            self.objective = config['objective']
            self.flux_bounds = config.get('flux_bounds', {})
            self.molecular_weights = config.get('molecular_weights', {})
            self.exchange_bounds = config.get('exchange_bounds', {})
            self.default_upper_bound = config.get('default_upper_bound', DEFAULT_UPPER_BOUND)
            self.flux_scaling = config.get('flux_scaling', 1)

            self.model = build_model(
                self.stoichiometry,
                self.reversible,
                self.objective,
                self.external_molecules,
                self.default_upper_bound)

            # apply constraints
            self.constrain_reaction_bounds(self.flux_bounds)
            self.set_exchange_bounds()

        self.exchange_bounds_keys = list(self.exchange_bounds.keys())

        # get minimal external state
        # TODO -- make sure that scaling is accounted for
        max_growth = self.model.slim_optimize()
        max_exchange = minimal_medium(self.model, max_growth)
        self.minimal_external = {ex[len(EXTERNAL_PREFIX):len(ex)]: value
            for ex, value in max_exchange.items()}

        # initialize solution
        self.solution = self.model.optimize()

    def set_exchange_bounds(self, bounds={}):
        '''
        apply new_bounds for the defined molecules.
        reset unincluded molecules to their exchange_bounds.
        '''
        for external_mol, level in self.exchange_bounds.items():
            reaction = self.model.reactions.get_by_id(EXTERNAL_PREFIX + external_mol)

            if external_mol in bounds:
                level = bounds[external_mol] / self.flux_scaling

            if type(level) is list:
                reaction.upper_bound = level[1]
                reaction.lower_bound = level[0]
            elif isinstance(level, int) or isinstance(level, float):
                # reaction.upper_bound = level
                reaction.lower_bound = level

    def constrain_flux(self, bounds={}):
        '''add externally imposed constraints'''
        for reaction_id, bound in bounds.items():
            reaction = self.model.reactions.get_by_id(reaction_id)
            scaled_level = bound / self.flux_scaling

            if EXTERNAL_PREFIX in reaction_id and \
                any(substring in reaction_id for substring in self.exchange_bounds_keys):
                # exchanges use reverse flux
                scaled_level *= -1

            if reaction_id in self.tolerance:
                # use configured tolerance
                lower_tolerance, upper_tolerance = self.tolerance[reaction_id]
                reaction.upper_bound = upper_tolerance * scaled_level
                reaction.lower_bound = lower_tolerance * scaled_level
            else:
                # use default tolerance
                if bound >= 0:
                    reaction.upper_bound = self.default_tolerance[1] * scaled_level
                    reaction.lower_bound = self.default_tolerance[0] * scaled_level
                else:
                    reaction.upper_bound = self.default_tolerance[0] * scaled_level
                    reaction.lower_bound = self.default_tolerance[1] * scaled_level

    def constrain_reaction_bounds(self, bounds={}):
        reactions = self.get_reactions(list(bounds.keys()))
        for reaction_id, bound in bounds.items():
            reaction = reactions[reaction_id]
            scaled_bound = [b / self.flux_scaling for b in bound]
            reaction.lower_bound, reaction.upper_bound = scaled_bound

    def regulate_flux(self, reactions):
        '''regulate flux based on True/False activity values for each id in reactions dictionary'''
        for reaction_id, activity in reactions.items():
            reaction = self.model.reactions.get_by_id(reaction_id)

            if not activity:
                # no activity. reaction flux set to 0
                reaction.upper_bound = 0.0
                reaction.lower_bound = 0.0
            elif activity and reaction.bounds == (0.0, 0.0):
                # if new bounds need to be set
                if reaction_id in self.flux_bounds:
                    bounds = self.flux_bounds[reaction_id]
                    scaled_bounds = [b / self.flux_scaling for b in bounds]
                    reaction.lower_bound, reaction.upper_bound = scaled_bounds
                elif reaction_id in self.reversible:
                    reaction.upper_bound = self.default_upper_bound / self.flux_scaling
                    reaction.lower_bound = -self.default_upper_bound / self.flux_scaling
                else:
                    # set bounds based on default
                    reaction.upper_bound = self.default_upper_bound / self.flux_scaling
                    reaction.lower_bound = 0.0

    def objective_value(self):
        if self.solution:
            objective_value = self.solution.objective_value * self.flux_scaling
            return objective_value
        else:
            return float('nan')

    def optimize(self):

        if self.moma:
            self.solution = cobra.flux_analysis.moma(self.model, solution=self.solution)
        else:
            self.solution = self.model.optimize()

        return self.objective_value()

    def external_reactions(self):
        return [
            EXTERNAL_PREFIX + molecule
            for molecule in self.external_molecules]

    def internal_reactions(self):
        all_reactions = set(self.reaction_ids())
        return all_reactions - set(self.external_reactions())

    def read_fluxes(self, molecules):
        return {
            molecule: self.solution.fluxes[molecule] * self.flux_scaling
            for molecule in molecules}

    def read_internal_fluxes(self):
        return self.read_fluxes(self.internal_reactions())

    def read_exchange_reactions(self):
        '''leaves the prefix in exchange reactions keys'''
        return self.read_fluxes(self.external_reactions())

    def read_exchange_fluxes(self):
        '''removes the prefix from exchange reactions keys, leaving only the external molecule id'''
        external = self.external_reactions()
        levels = self.read_fluxes(external)
        return {
            molecule[len(EXTERNAL_PREFIX):len(molecule)]: level
            for molecule, level in levels.items()}

    def reaction_ids(self):
        return [reaction.id for reaction in self.model.reactions]

    def get_reactions(self, reactions=[]):
        if not reactions:
            reactions = self.reaction_ids()

        return {
            reaction: self.model.reactions.get_by_id(reaction)
            for reaction in reactions}

    def get_reaction_bounds(self, reactions=[]):
        return {
            reaction_key: (reaction.lower_bound, reaction.upper_bound)
            for reaction_key, reaction in self.get_reactions(reactions).items()}



def test_minimal():
    stoichiometry = {
        'R1': {'A': -1, 'B': 1},
        'EB': {'B': -1}}

    objective = {'EB': 1.0}

    external_molecules = ['A']

    initial_state = {
        'A': 5}

    fba = CobraFBA({
        'stoichiometry': stoichiometry,
        'reversible': list(stoichiometry.keys()),
        'objective': objective,
        'external_molecules': external_molecules,
        'initial_state': initial_state,
        })

    fba.set_exchange_bounds(initial_state)

    return fba

def test_fba():
    stoichiometry = {
        'R1': {'A': -1, 'ATP': -1, 'B': 1},
        'R2a': {'B': -1, 'ATP': 2, 'NADH': 2, 'C': 1},
        'R2b': {'C': -1, 'ATP': -2, 'NADH': -2, 'B': 1},
        'R3': {'B': -1, 'F': 1},
        'R4': {'C': -1, 'G': 1},
        'R5': {'G': -1, 'C': 0.8, 'NADH': 2},
        'R6': {'C': -1, 'ATP': 2, 'D': 3},
        'R7': {'C': -1, 'NADH': -4, 'E': 3},
        'R8a': {'G': -1, 'ATP': -1, 'NADH': -2, 'H': 1},
        'R8b': {'G': 1, 'ATP': 1, 'NADH': 2, 'H': -1},
        'Rres': {'NADH': -1, 'O2': -1, 'ATP': 1},
        'v_biomass': {'C': -1, 'F': -1, 'H': -1, 'ATP': -10, 'BIOMASS': 1}
    }

    external_molecules = ['A', 'F', 'D', 'E', 'H', 'O2', 'BIOMASS']

    objective = {'v_biomass': 1.0}

    initial_state = {
        'internal': {
            'mass': 1339,
            'volume': 1E-15},
        'external': {
            'A': 21.0,
            'F': 5.0,
            'D': 12.0,
            'E': 12.0,
            'H': 5.0,
            'O2': 100.0}}

    fba = CobraFBA({
        'stoichiometry': stoichiometry,
        'reversible': list(stoichiometry.keys()),
        'objective': objective,
        'external_molecules': external_molecules})

    fba.set_exchange_bounds(initial_state['external'])

    return fba

class JsonFBA(object):
    def __init__(self, path):
        self.model = cobra.io.load_json_model(path)

def test_canonical():
    fba = JsonFBA('vivarium_cell/bigg_models/e_coli_core.json')
    return fba

def test_demo():
    model = Model('example_model')

    reaction = Reaction('3OAS140')
    reaction.name = '3 oxoacyl acyl carrier protein synthase n C140 '
    reaction.subsystem = 'Cell Envelope Biosynthesis'
    reaction.lower_bound = 0.  # This is the default
    reaction.upper_bound = 1000.  # This is the default

    ACP_c = Metabolite(
        'ACP_c',
        formula='C11H21N2O7PRS',
        name='acyl-carrier-protein',
        compartment='c')
    omrsACP_c = Metabolite(
        '3omrsACP_c',
        formula='C25H45N2O9PRS',
        name='3-Oxotetradecanoyl-acyl-carrier-protein',
        compartment='c')
    co2_c = Metabolite('co2_c', formula='CO2', name='CO2', compartment='c')
    malACP_c = Metabolite(
        'malACP_c',
        formula='C14H22N2O10PRS',
        name='Malonyl-acyl-carrier-protein',
        compartment='c')
    h_c = Metabolite('h_c', formula='H', name='H', compartment='c')
    ddcaACP_c = Metabolite(
        'ddcaACP_c',
        formula='C23H43N2O8PRS',
        name='Dodecanoyl-ACP-n-C120ACP',
        compartment='c')

    reaction.add_metabolites({
        malACP_c: -1.0,
        h_c: -1.0,
        ddcaACP_c: -1.0,
        co2_c: 1.0,
        ACP_c: 1.0,
        omrsACP_c: 1.0
    })

    print(reaction.reaction)  # This gives a string representation of the reaction

    reaction.gene_reaction_rule = '( STM2378 or STM1197 )'
    print(reaction.genes)

    model.add_reactions([reaction])

    # Now there are things in the model
    print('%i reaction' % len(model.reactions))
    print('%i metabolites' % len(model.metabolites))
    print('%i genes' % len(model.genes))

    # Iterate through the the objects in the model
    print('Reactions')
    print('---------')
    for x in model.reactions:
        print('%s : %s' % (x.id, x.reaction))

    print('')
    print('Metabolites')
    print('-----------')
    for x in model.metabolites:
        print('%9s : %s' % (x.id, x.formula))

    print('')
    print('Genes')
    print('-----')
    for x in model.genes:
        associated_ids = (i.id for i in x.reactions)
        print('%s is associated with reactions: %s' %
              (x.id, '{' + ', '.join(associated_ids) + '}'))

    model.objective = '3OAS140'

    print(model.objective.expression)
    print(model.objective.direction)

    class DemoFBA(object):
        def __init__(self, model):
            self.model = model

    return DemoFBA(model)


if __name__ == '__main__':
    fba = test_fba()
    # fba = test_minimal()
    # fba = test_canonical()
    # fba = test_demo()
    # fba = test_test()

    # cobra.io.save_json_model(fba.model, 'demo_model.json')
    print('MODEL: {}'.format(fba.model))
    print('REACTIONS: {}'.format(fba.model.reactions))
    print('METABOLITES: {}'.format(fba.model.metabolites))
    print('GENES: {}'.format(fba.model.genes))
    print('COMPARTMENTS: {}'.format(fba.model.compartments))
    print('SOLVER: {}'.format(fba.model.solver))
    print('EXPRESSION: {}'.format(fba.model.objective.expression))

    print(fba.optimize())
    print(fba.model.summary())
    print('internal: {}'.format(fba.internal_reactions()))
    print('external: {}'.format(fba.external_reactions()))
    print(fba.reaction_ids())
    print(fba.get_reactions())
    print(fba.get_reaction_bounds())
    print(fba.read_exchange_fluxes())



from __future__ import absolute_import, division, print_function

import os

from vivarium.core.process import Generator
from vivarium.core.composition import (
    compartment_in_experiment,
    COMPARTMENT_OUT_DIR,
)

# processes
from vivarium_cell.processes.multibody_physics import (
    Multibody,
    agent_body_config,
)
from vivarium_cell.plots.multibody_physics import plot_snapshots
from vivarium_cell.processes.static_field import StaticField

NAME = 'static_lattice'


class StaticLattice(Generator):

    defaults = {
        'multibody': {
            'bounds': [10, 10],
            'agents': {}
        },
        'field': {
            'molecules': ['glc'],
            'n_bins': [10, 10],
            'bounds': [10, 10],
        }
    }

    def __init__(self, config):
        super(StaticLattice, self).__init__(config)

    def generate_processes(self, config=None):
        multibody = Multibody(config['multibody'])
        field = StaticField(config['field'])

        return {
            'multibody': multibody,
            'field': field}

    def generate_topology(self, config=None):
        return {
            'multibody': {
                'agents': ('agents',)},
            'field': {
                'agents': ('agents',)}}


def get_static_lattice_config(config={}):
    bounds = config.get('bounds', [25, 25])
    molecules = config.get('molecules', ['glc'])
    n_bins = config.get('n_bins', tuple(bounds))
    center = config.get('center', [0.5, 0.5])
    deviation = config.get('deviation', 5)
    n_agents = config.get('n_agents', 1)
    agent_ids = [str(agent_id) for agent_id in range(n_agents)]

    # multibody config
    mbp_config = {
        # 'animate': True,
        'jitter_force': 1e2,
        'bounds': bounds}
    body_config = {
        'bounds': bounds,
        'agent_ids': agent_ids}
    mbp_config['agents'] = agent_body_config(body_config)

    # field config
    field_config = {
        'molecules': molecules,
        'n_bins': n_bins,
        'bounds': bounds,
        # 'agents': agents,
        'gradient': {
            'type': 'exponential',
            'molecules': {
                mol_id: {
                    'center': [0.0, 0.0],
                    'base': 1 + 1e-1}
                for mol_id in molecules}},
        # 'initial_state': {
        #     mol_id: np.ones((n_bins[0], n_bins[1]))
        #     for mol_id in molecules}
    }

    return {
        'bounds': bounds,
        'multibody': mbp_config,
        'field': field_config}

def test_static_lattice(config=get_static_lattice_config(), end_time=10):

    # configure the compartment
    compartment = StaticLattice(config)

    # configure experiment
    experiment_settings = {
        'compartment': config}
    experiment = compartment_in_experiment(
        compartment,
        experiment_settings)

    # run experiment
    timestep = 1
    time = 0
    while time < end_time:
        experiment.update(timestep)
        time += timestep
    return experiment.emitter.get_data()



if __name__ == '__main__':
    out_dir = os.path.join(COMPARTMENT_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    config = get_static_lattice_config()
    data = test_static_lattice(config, 40)




'''
===============================
Model of Conditional Cell Death
===============================

Fully-capitalized words and phrases have the meanings specified in
:rfc:`2119`.

This module holds machinery for modeling cell death that is conditioned
on the state of the cell. This machinery consists of *death detector
classes* and *death process classes*.

----------------------
Death Detector Classes
----------------------

Death detector classes encode a model (which may be configured upon
instantiation) for when a cell should die. These classes can be
instantiated to give death detectors, which are used by the death
processes we describe below.

Death detector classes MUST subclass :py:class:`DetectorInterface` and
implement :py:meth:`DetectorInterface.check_can_survive` as specified in
its documentation.

---------------------
Death Process Classes
---------------------

During a simulation, death is executed by a death :term:`process`, whose
model is declared in a death :term:`process class`. Each of these
process classes SHOULD use a death detector, as detailed above, to
determine when the cell should die. The mechanism by which the cell's
death is modeled depends on the form of death being modeled. For an
example, see :py:class:`DeathFreezeState`.
'''

from __future__ import absolute_import, division, print_function

import os

from vivarium.core.composition import (
    plot_simulation_output,
    simulate_compartment_in_experiment,
    PROCESS_OUT_DIR,
)
from vivarium.core.process import Process, Generator

TOY_ANTIBIOTIC_THRESHOLD = 5.0
TOY_INJECTION_RATE = 2.0


class DetectorInterface(object):
    '''Interface that MUST be subclassed by all death detectors

    Each subclass SHOULD check for a condition that might kill the cell.

    For an example of a death detector class, see
    :py:class:`AntibioticDetector`.
    '''

    def __init__(self):
        self.needed_state_keys = {}

    def check_can_survive(self, states):
        '''Check whether the current state is survivable by the cell

        Each subclass MUST implement this method. The implementation
        SHOULD check whether the cell can survive given its current
        state.

        Arguments:
            states (dict): The states of each port in the cell, as a
                dictionary.

        Returns:
            bool: True if the cell can survive, False if it cannot.
        '''
        raise NotImplementedError(
            'Detector should implement check_can_survive')


class AntibioticDetector(DetectorInterface):

    def __init__(
        self, antibiotic_threshold=0.9, antibiotic_key='antibiotic'
    ):
        '''Death detector for antibiotics

        Checks whether the cell can survive the current internal
        antibiotic concentrations.

        Arguments:
            antibiotic_threshold (float): The maximum internal
                antibiotic concentration the cell can survive.
            antibiotic_key (str): The name of the variable storing
                the cell's internal antibiotic concentration.
        '''
        super(AntibioticDetector, self).__init__()
        self.threshold = antibiotic_threshold
        self.key = antibiotic_key
        self.needed_state_keys.setdefault(
            'internal', set()).add(antibiotic_key)

    def check_can_survive(self, states):
        '''Checks if the current antibiotic concentration is survivable

        The internal antibiotic concentration MUST be stored in a
        variable of a port named ``internal``.

        Returns:
            bool: False if the antibiotic concentration is strictly
                greater than the the threshold. True otherwise.
        '''
        concentration = states['internal'][self.key]
        if concentration > self.threshold:
            return False
        return True


#: Map from detector class names to detector classes
DETECTOR_CLASSES = {
    'antibiotic': AntibioticDetector,
}


class DeathFreezeState(Process):

    name = 'death'

    def __init__(self, initial_parameters=None):
        '''Model Death by Removing Processes

        This process class models death by, with a few exceptions,
        freezing the internal state of the cell. We implement this by
        removing from this process's :term:`compartment` all processes,
        specified with the ``targets`` configuration.

        Configuration:

        * **``detectors``**: A list of the names of the detector classes
          to include. Death will be triggered if any one of these
          triggers death. Names are specified in
          :py:const:`DETECTOR_CLASSES`.
        * **``targets``**: A list of the names of the processes
          that will be removed when the cell dies. The names are
          specified in the compartment's :term:`topology`.

        :term:`Ports`:

        * **``internal``**: The internal state of the cell.
        * **``global``**: Should be linked to the ``global``
          :term:`store`.
        * **``processes``**: Should be linked to the store that has
          the processes as children.
        '''
        if initial_parameters is None:
            initial_parameters = {}
        self.detectors = [
            DETECTOR_CLASSES[name](**config_dict)
            for name, config_dict in initial_parameters.get(
                'detectors', {}).items()
        ]
        # List of names of processes that will be removed upon death
        self.targets = initial_parameters.get('targets', [])

        super(DeathFreezeState, self).__init__(initial_parameters)

    def ports_schema(self):

        schema = {'global': {}}

        # global
        schema['global']['dead'] = {
            '_default': 0,
            '_emit': True,
            '_updater': 'set'
        }
        schema['processes'] = {
            target: {
                '_default': None
            }
            for target in self.targets
        }

        # detector ports
        for detector in self.detectors:
            needed_keys = detector.needed_state_keys
            for port, states in needed_keys.items():
                if port not in schema:
                    schema[port] = {}
                for state in states:
                    schema[port][state] = {'_default': 0}

        return schema

    def next_update(self, timestep, states):
        '''If any detector triggers death, kill the cell

        When we kill the cell, we convey this by setting the ``dead``
        variable in the ``global`` port to ``1`` instead of its default
        ``0``.
        '''
        for detector in self.detectors:
            if not detector.check_can_survive(states):
                # kill the cell
                update = {
                    'global': {
                        'dead': 1,
                    },
                    'processes': {
                        '_delete': [
                            (target,)
                            for target in self.targets
                        ],
                    },
                }
                return update
        return {}


class ToyAntibioticInjector(Process):
    name = 'toy_antibiotic_injector'

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}
        self.injection_rate = initial_parameters.get(
            'injection_rate', 1.0)
        self.antibiotic_name = initial_parameters.get(
            'antibiotic_name', 'antibiotic')
        super(ToyAntibioticInjector, self).__init__(initial_parameters)

    def ports_schema(self):
        return {
            'internal': {
                self.antibiotic_name: {
                    '_default': 0.0,
                    '_emit': True}}}

    def next_update(self, timestep, states):
        delta = timestep * self.injection_rate
        return {'internal': {self.antibiotic_name: delta}}


class ToyDeath(Generator):

    def generate_processes(self, config):
        death_parameters = {
            'detectors': {
                'antibiotic': {
                    'antibiotic_threshold': TOY_ANTIBIOTIC_THRESHOLD,
                }
            },
            'targets': ['injector', 'death'],
        }
        death_process = DeathFreezeState(death_parameters)
        injector_parameters = {
            'injection_rate': TOY_INJECTION_RATE,
        }
        injector_process = ToyAntibioticInjector(injector_parameters)
        enduring_parameters = {
            'injection_rate': TOY_INJECTION_RATE,
            'antibiotic_name': 'enduring_antibiotic'
        }
        enduring_process = ToyAntibioticInjector(enduring_parameters)

        return {
            'death': death_process,
            'injector': injector_process,
            'enduring_injector': enduring_process,
        }

    def generate_topology(self, config):
        return {
            'death': {
                'internal': ('cell',),
                'global': ('global',),
                'processes': tuple(),
            },
            'injector': {
                'internal': ('cell',),
            },
            'enduring_injector': {
                'internal': ('cell',),
            },
        }



def test_death_freeze_state(end_time=10, asserts=True):
    toy_death_compartment = ToyDeath({})

    init_state = {
        'cell': {
            'antibiotic': 0.0,
            'enduring_antibiotic': 0.0},
        'global': {
            'dead': 0}}

    settings = {
        'total_time': end_time,
        'initial_state': init_state}
    saved_states = simulate_compartment_in_experiment(
        toy_death_compartment,
        settings)

    if asserts:
        # Add 1 because dies when antibiotic strictly above threshold
        expected_death = 1 + TOY_ANTIBIOTIC_THRESHOLD // TOY_INJECTION_RATE
        expected_saved_states = {
            'cell': {
                'antibiotic': [],
                'enduring_antibiotic': [],
            },
            'global': {
                'dead': [],
            },
            'time': [],
        }
        for time in range(end_time + 1):
            expected_saved_states['cell']['antibiotic'].append(
                time * TOY_INJECTION_RATE
                if time <= expected_death
                # Add one because death will only be detected
                # the iteration after antibiotic above
                # threshold. This happens because death and
                # injector run "concurrently" in the composite,
                # so their updates are applied after both have
                # finished.
                else (expected_death + 1) * TOY_INJECTION_RATE
            )
            expected_saved_states['cell']['enduring_antibiotic'].append(
                time * TOY_INJECTION_RATE)
            expected_saved_states['global']['dead'].append(
                0 if time <= expected_death else 1)
            expected_saved_states['time'].append(float(time))

        assert expected_saved_states == saved_states

    return saved_states


def plot_death_freeze_state_test():
    out_dir = os.path.join(PROCESS_OUT_DIR, 'death_freeze_state')
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    timeseries = test_death_freeze_state(asserts=False)
    plot_settings = {}
    plot_simulation_output(timeseries, plot_settings, out_dir)


if __name__ == '__main__':
    plot_death_freeze_state_test()



from __future__ import absolute_import, division, print_function

import argparse
import csv
import os

from vivarium_cell.plots.multibody_physics import (
    plot_snapshots,
    plot_tags,
)
from vivarium.core.composition import plot_agents_multigen
from vivarium.core.emitter import (
    get_atlas_client,
    get_local_client,
    data_from_database,
    SECRETS_PATH,
    timeseries_from_data,
    path_timeseries_from_embedded_timeseries,
)
from vivarium_cell.plots.colonies import plot_colony_metrics


OUT_DIR = 'out'


class Analyzer:

    def __init__(
        self,
        snapshots_config=None,
        tags_config=None,
        timeseries_config=None,
    ):
        self.parser = self._get_parser()
        if snapshots_config is None:
            snapshots_config = {}
        if tags_config is None:
            tags_config = {}
        if timeseries_config is None:
            timeseries_config = {}
        self.snapshots_config = snapshots_config
        self.tags_config = tags_config
        self.timeseries_config = timeseries_config

    def run(self):
        args = self.parser.parse_args()
        self.plot(args)

    @staticmethod
    def get_data(args, experiment_id):
        if args.atlas:
            client = get_atlas_client(SECRETS_PATH)
        else:
            client = get_local_client(
                args.host, args.port, args.database_name)
        data, environment_config = data_from_database(
            experiment_id, client)
        del data[0]
        return data, environment_config

    @staticmethod
    def plot_snapshots(data, environment_config, out_dir, settings):
        snapshots_data = Analyzer.format_data_for_snapshots(
            data, environment_config)
        plot_config = {
            'out_dir': out_dir,
        }
        plot_config.update(settings)
        plot_snapshots(snapshots_data, plot_config)

    @staticmethod
    def format_data_for_snapshots(data, environment_config):
        agents = {
            time: timepoint['agents']
            for time, timepoint in data.items()
        }
        fields = {
            time: timepoint['fields']
            for time, timepoint in data.items()
        }
        snapshots_data = {
            'agents': agents,
            'fields': fields,
            'config': environment_config,
        }
        return snapshots_data

    @staticmethod
    def plot_tags(
        data, environment_config, tagged_molecules, out_dir, settings
    ):
        tags_data = Analyzer.format_data_for_tags(
            data, environment_config)
        plot_config = {
            'out_dir': out_dir,
            'tagged_molecules': tagged_molecules,
        }
        plot_config.update(settings)
        plot_tags(tags_data, plot_config)

    @staticmethod
    def format_data_for_tags(data, environment_config):
        agents = {
            time: timepoint['agents']
            for time, timepoint in data.items()
        }
        tags_data = {
            'agents': agents,
            'config': environment_config,
        }
        return tags_data

    @staticmethod
    def plot_timeseries(data, out_dir, settings):
        plot_settings = {
            'agents_key': 'agents',
            'title_size': 10,
            'tick_label_size': 10,
        }
        plot_settings.update(settings)
        plot_agents_multigen(data, plot_settings, out_dir)

    @staticmethod
    def plot_colony_metrics(data, out_dir):
        path_ts = Analyzer.format_data_for_colony_metrics(data)
        fig = plot_colony_metrics(path_ts)
        fig.savefig(os.path.join(out_dir, 'colonies'))

    @staticmethod
    def format_data_for_colony_metrics(data):
        embedded_ts = timeseries_from_data(data)
        colony_metrics_ts = embedded_ts['colony_global']
        colony_metrics_ts['time'] = embedded_ts['time']
        path_ts = path_timeseries_from_embedded_timeseries(
            colony_metrics_ts)
        return path_ts

    def plot(self, args):
        data, environment_config = Analyzer.get_data(
            args, args.experiment_id)
        out_dir = os.path.join(OUT_DIR, args.experiment_id)
        if os.path.exists(out_dir):
            if not args.force:
                raise IOError('Directory {} already exists'.format(out_dir))
        else:
            os.makedirs(out_dir)

        if args.snapshots:
            Analyzer.plot_snapshots(
                data, environment_config, out_dir,
                self.snapshots_config
            )
        if args.tags is not None:
            with open(args.tags, 'r') as f:
                reader = csv.reader(f)
                tagged_molecules = [
                    tuple(path) for path in reader
                ]
            Analyzer.plot_tags(
                data, environment_config, tagged_molecules, out_dir,
                self.tags_config
            )
        if args.timeseries:
            Analyzer.plot_timeseries(
                data, out_dir, self.timeseries_config)
        if args.colony_metrics:
            Analyzer.plot_colony_metrics(data, out_dir)

    def _get_parser(self):
        parser = argparse.ArgumentParser()
        Analyzer.add_connection_args(parser)
        parser.add_argument(
            'experiment_id',
            help='Experiment ID as recorded in the database',
        )
        parser.add_argument(
            '--snapshots', '-s',
            action='store_true',
            default=False,
            help='Plot snapshots',
        )
        parser.add_argument(
            '--tags', '-g',
            default=None,
            help=(
                'A path to a CSV file that lists the tagged molecules to '
                'plot. The first column should contain the name of the store '
                'under each agent boundary where the molecule is reported, '
                'and the second column should contain the name of the '
                'molecule. Setting this parameter causes a plot of the tagged '
                'molecules to be produced.'
            ),
        )
        parser.add_argument(
            '--timeseries', '-t',
            action='store_true',
            default=False,
            help='Generate line plot for each variable over time',
        )
        parser.add_argument(
            '--colony_metrics', '-c',
            action='store_true',
            default=False,
            help='Plot colony metrics',
        )
        parser.add_argument(
            '--force', '-f',
            action='store_true',
            default=False,
            help=(
                'Write plots even if output directory already exists. This '
                'could overwrite your existing plots'
            ),
        )
        return parser

    @staticmethod
    def add_connection_args(parser):
        parser.add_argument(
            '--atlas', '-a',
            action='store_true',
            default=False,
            help=(
                'Read data from an mongoDB Atlas instead of a local mongoDB. '
                'Credentials, cluster subdomain, and database name should be '
                'specified in {}.'.format(SECRETS_PATH)
            )
        )
        parser.add_argument(
            '--port', '-p',
            default=27017,
            type=int,
            help=(
                'Port at which to access local mongoDB instance. '
                'Defaults to "27017".'
            ),
        )
        parser.add_argument(
            '--host', '-o',
            default='localhost',
            type=str,
            help=(
                'Host at which to access local mongoDB instance. '
                'Defaults to "localhost".'
            ),
        )
        parser.add_argument(
            '--database_name', '-d',
            default='simulations',
            type=str,
            help=(
                'Name of database on local mongoDB instance to read from. '
                'Defaults to "simulations".'
            )
        )


if __name__ == '__main__':
    analyzer = Analyzer()
    analyzer.run()



"""
================================================
Toy Injected Glucose Phosphorylation Compartment
================================================

This is a toy example referenced in the documentation.
"""

from vivarium.core.experiment import Experiment
from vivarium.core.process import Generator
from vivarium.library.pretty import format_dict

from vivarium_cell.processes.glucose_phosphorylation import GlucosePhosphorylation
from vivarium_cell.processes.injector import Injector


class InjectedGlcPhosphorylation(Generator):

    defaults = {
        'glucose_phosphorylation': {
            'k_cat': 1e-2,
        },
        'injector': {
            'substrate_rate_map': {
                'GLC': 1e-4,
                'ATP': 1e-3,
            },
        },
    }

    def __init__(self, config):
        super(InjectedGlcPhosphorylation, self).__init__(config)

    def generate_processes(self, config):
        injector = Injector(self.config['injector'])
        glucose_phosphorylation = GlucosePhosphorylation(
            self.config['glucose_phosphorylation'])

        return {
            'injector': injector,
            'glucose_phosphorylation': glucose_phosphorylation,
        }

    def generate_topology(self, config):
        return {
            'injector': {
                'internal': ('cell', ),
            },
            'glucose_phosphorylation': {
                'cytoplasm': ('cell', ),
                'nucleoside_phosphates': ('cell', ),
                'global': ('global', ),
            },
        }



from __future__ import absolute_import, division, print_function

import pprint

from arpeggio import Optional, ZeroOrMore, OneOrMore, EOF, ParserPython, Kwd, RegExMatch


pp = pprint.PrettyPrinter(indent=4)

def convert_number(s):
    try:
        return float(s)
    except:
        return None


# each function returns a tuple
def symbol(): return RegExMatch(r'[a-zA-Z0-9.\-\_]+')  # TODO -- surplus can be evaluated if there is a threshold, ignored for now
def tuple_key(): return Kwd("("), symbol, Kwd(","), symbol, Kwd(")")
def operator(): return [Kwd('>'), Kwd('<')]
def compare(): return [symbol, tuple_key], Optional(operator, symbol)
def group(): return Kwd("["), logic, Kwd("]")
def term(): return Optional(Kwd("not")), [compare, group]
def logic(): return term, ZeroOrMore([Kwd("and"), Kwd("or")], term)
def rule(): return Kwd("if"), logic, EOF

def evaluate_symbol(tree, state):
    symbol = tree
    value = state.get(symbol.value)
    if value is None:
        # symbol is a value, used by compare
        value = convert_number(symbol.value)

    return value

def evaluate_tuple_key(tree, state):
    tuple_key = (tree[1].value, tree[3].value)
    value = state.get(tuple_key)

    return value

def evaluate_compare(tree, state):
    if tree[0].rule_name == 'tuple_key':
        first = evaluate_tuple_key(tree[0], state)
    else:
        first = evaluate_symbol(tree[0], state)

    if len(tree) == 1:
        if isinstance(first, int) or isinstance(first, float):
            # if numeric state, evaluate whether it is present or not
            return first > 0
        else:
            return first
    else:
        operator = tree[1]
        last = evaluate_symbol(tree[2], state)
        if operator.value == '<':
            return first < last
        else:
            return first > last

def evaluate_group(tree, state):
    logic = tree[1]
    return evaluate_logic(logic, state)

def evaluate_term(tree, state):
    invert = False
    value = False

    if tree[0].value == 'not':
        invert = True
        tree = tree[1:]

    if tree[0].rule_name == 'group':
        value = evaluate_group(tree[0], state)
    elif tree[0].rule_name == 'compare':
        value = evaluate_compare(tree[0], state)

    if invert:
        value = not value

    return value

def evaluate_logic(tree, state):
    head = evaluate_term(tree[0], state)

    if len(tree) > 1:
        tail = evaluate_logic(tree[2:], state)
        operation = tree[1].value

        if operation == 'and':
            head = head and tail
        elif operation == 'or':
            head = head or tail

    return head

def evaluate_rule(tree, state):
    return evaluate_logic(tree[1], state)


# make parser based on "rule", defined in grammar above
rule_parser = ParserPython(rule)

def build_rule(expression):
    # type: (str) -> Callable[Dict[str, bool], bool]

    '''
    Accepts a string representing a logical statement about the presence or absence of
    various molecular entities relevant to regulation, and returns a function that
    evaluates that logic with respect to actual values for the various symbols. 
    '''
    tree = rule_parser.parse(expression)

    def logic(state):
        return evaluate_rule(tree, state)

    return logic

def test_arpeggio():

    # test logic with sets
    test = "if not [GLCxt or LCTSxt or RUBxt] and FNR and not GlpR"
    state_false = {'GLCxt': True, 'LCTSxt': False, 'RUBxt': True, 'FNR': True, 'GlpR': False}
    state_true = {'GLCxt': False, 'LCTSxt': False, 'RUBxt': False, 'FNR': True, 'GlpR': False}
    run_rule = build_rule(test)
    assert run_rule(state_false) == False
    assert run_rule(state_true) == True

    # test logic with thresholds
    test = "if not glc > 0.1"
    state_false = {'glc': 0.2}
    state_true = {'glc': 0.01}
    run_rule = build_rule(test)
    assert run_rule(state_false) == False
    assert run_rule(state_true) == True

    # test thresholds in sets with numeric states
    test = "if not [FDP > 10 and F6P]"  # same as "if not [FDP > 10 and F6P > 0]"
    state_false = {'FDP': 20, 'F6P': 10}
    state_true = {'FDP': 5, 'F6P': 0}
    run_rule = build_rule(test)
    assert run_rule(state_false) == False
    assert run_rule(state_true) == True

    # test tuple keys
    test = "if not [(external, glc) > 0.1 and (external, lct) < 0.1]"
    state_false = {
        ('external', 'glc'): 0.2,
        ('external', 'lct'): 0.05}
    state_true = {
        ('external', 'glc'): 0.01,
        ('external', 'lct'): 0.5}
    run_rule = build_rule(test)
    assert run_rule(state_false) == False
    assert run_rule(state_true) == True

    print('test passed!')

    return run_rule


if __name__ == '__main__':
    test_arpeggio()



'''
===============
Diffusion Field
===============
'''

from __future__ import absolute_import, division, print_function

import sys
import os
import argparse

import numpy as np
from scipy import constants
from scipy.ndimage import convolve

from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process,
    PROCESS_OUT_DIR
)
from vivarium.library.units import units

from vivarium_cell.library.lattice_utils import (
    count_to_concentration,
    get_bin_site,
    get_bin_volume,
)
from vivarium_cell.plots.multibody_physics import plot_snapshots

NAME = 'diffusion_field'

# laplacian kernel for diffusion
LAPLACIAN_2D = np.array([[0.0, 1.0, 0.0], [1.0, -4.0, 1.0], [0.0, 1.0, 0.0]])
AVOGADRO = constants.N_A


def gaussian(deviation, distance):
    return np.exp(-np.power(distance, 2.) / (2 * np.power(deviation, 2.)))

def make_gradient(gradient, n_bins, size):
    '''Create a gradient from a configuration

    **Uniform**

    A uniform gradient fills the field evenly with each molecule, at
    the concentrations specified.

    Example configuration:

    .. code-block:: python

        'gradient': {
            'type': 'uniform',
            'molecules': {
                'mol_id1': 1.0,
                'mol_id2': 2.0
            }},

    **Gaussian**

    A gaussian gradient multiplies the base concentration of the given
    molecule by a gaussian function of distance from center and
    deviation. Distance is scaled by 1/1000 from microns to millimeters.

    Example configuration:

    .. code-block:: python

        'gradient': {
            'type': 'gaussian',
            'molecules': {
                'mol_id1':{
                    'center': [0.25, 0.5],
                    'deviation': 30},
                'mol_id2': {
                    'center': [0.75, 0.5],
                    'deviation': 30}
            }},

    **Linear**

    A linear gradient sets a site's concentration (c) of the given
    molecule as a function of distance (d) from center and slope (b),
    and base concentration (a). Distance is scaled by 1/1000 from
    microns to millimeters.

    .. math::
        c = a + b * d

    Example configuration:

    .. code-block:: python

        'gradient': {
            'type': 'linear',
            'molecules': {
                'mol_id1':{
                    'center': [0.0, 0.0],
                    'base': 0.1,
                    'slope': -10},
                'mol_id2': {
                    'center': [1.0, 1.0],
                    'base': 0.1,
                    'slope': -5}
            }},

    **Exponential**

    An exponential gradient sets a site's concentration (c) of the given
    molecule as a function of distance (d) from center, with parameters
    base (b) and scale (a). Distance is scaled by 1/1000 from microns to
    millimeters. Note: base > 1 makes concentrations increase from the
    center.

    .. math::

        c=a*b^d.

    Example configuration:

    .. code-block:: python

        'gradient': {
            'type': 'exponential',
            'molecules': {
                'mol_id1':{
                    'center': [0.0, 0.0],
                    'base': 1+2e-4,
                    'scale': 1.0},
                'mol_id2': {
                    'center': [1.0, 1.0],
                    'base': 1+2e-4,
                    'scale' : 0.1}
            }},

    Parameters:
        gradient: Configuration dictionary that includes the ``type``
            key to specify the type of gradient to make.
        n_bins: A list of two elements that specify the number of bins
            to have along each axis.
        size: A list of two elements that specifies the size of the
            environment.
    '''
    bins_x = n_bins[0]
    bins_y = n_bins[1]
    length_x = size[0]
    length_y = size[1]
    fields = {}

    if gradient.get('type') == 'gaussian':
        for molecule_id, specs in gradient['molecules'].items():
            field = np.ones((bins_x, bins_y), dtype=np.float64)
            center = [specs['center'][0] * length_x,
                      specs['center'][1] * length_y]
            deviation = specs['deviation']

            for x_bin in range(bins_x):
                for y_bin in range(bins_y):
                    # distance from middle of bin to center coordinates
                    dx = (x_bin + 0.5) * length_x / bins_x - center[0]
                    dy = (y_bin + 0.5) * length_y / bins_y - center[1]
                    distance = np.sqrt(dx ** 2 + dy ** 2)
                    scale = gaussian(deviation, (distance/1000))
                    # multiply gradient by scale
                    field[x_bin][y_bin] *= scale
            fields[molecule_id] = field

    elif gradient.get('type') == 'linear':
        for molecule_id, specs in gradient['molecules'].items():
            field = np.zeros((bins_x, bins_y), dtype=np.float64)
            center = [specs['center'][0] * length_x,
                      specs['center'][1] * length_y]
            base = specs.get('base', 0.0)
            slope = specs['slope']

            for x_bin in range(bins_x):
                for y_bin in range(bins_y):
                    dx = (x_bin + 0.5) * length_x / bins_x - center[0]
                    dy = (y_bin + 0.5) * length_y / bins_y - center[1]
                    distance = np.sqrt(dx ** 2 + dy ** 2)
                    field[x_bin][y_bin] += base + slope * (distance/1000)
            fields[molecule_id] = field

    elif gradient.get('type') == 'exponential':
        for molecule_id, specs in gradient['molecules'].items():
            field = np.zeros((bins_x, bins_y), dtype=np.float64)
            center = [specs['center'][0] * length_x,
                      specs['center'][1] * length_y]
            base = specs['base']
            scale = specs.get('scale', 1)

            for x_bin in range(bins_x):
                for y_bin in range(bins_y):
                    dx = (x_bin + 0.5) * length_x / bins_x - center[0]
                    dy = (y_bin + 0.5) * length_y / bins_y - center[1]
                    distance = np.sqrt(dx ** 2 + dy ** 2)
                    field[x_bin][y_bin] = scale * base ** (distance/1000)
            fields[molecule_id] = field

    elif gradient.get('type') == 'uniform':
        for molecule_id, fill_value in gradient['molecules'].items():
            fields[molecule_id] = np.full((bins_x, bins_y), fill_value, dtype=np.float64)

    return fields


class DiffusionField(Process):
    '''
    Diffusion in 2-dimensional fields of molecules with agent exchange

    Agent uptake and secretion occurs at agent locations.

    Notes:

    * Diffusion constant of glucose in 0.5 and 1.5 percent agarose gel
      is around :math:`6 * 10^{-10} \\frac{m^2}{s}` (Weng et al. 2005.
      Transport of glucose and poly(ethylene glycol)s in agarose gels).
    * Conversion to micrometers:
      :math:`6 * 10^{-10} \\frac{m^2}{s}=600 \\frac{micrometers^2}{s}`.
    '''

    name = NAME
    defaults = {
        'time_step': 1,
        'molecules': ['glc'],
        'initial_state': {},
        'n_bins': [10, 10],
        'bounds': [10, 10],
        'depth': 3000.0,  # um
        'diffusion': 5e-1,
        'gradient': {},
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        # initial state
        self.molecule_ids = initial_parameters.get('molecules', self.defaults['molecules'])
        self.initial_state = initial_parameters.get('initial_state', self.defaults['initial_state'])

        # parameters
        self.n_bins = initial_parameters.get('n_bins', self.defaults['n_bins'])
        self.bounds = initial_parameters.get('bounds', self.defaults['bounds'])
        depth = initial_parameters.get('depth', self.defaults['depth'])

        # diffusion
        diffusion = initial_parameters.get('diffusion', self.defaults['diffusion'])
        bins_x = self.n_bins[0]
        bins_y = self.n_bins[1]
        length_x = self.bounds[0]
        length_y = self.bounds[1]
        dx = length_x / bins_x
        dy = length_y / bins_y
        dx2 = dx * dy
        self.diffusion = diffusion / dx2
        self.diffusion_dt = 0.01
        # self.diffusion_dt = 0.5 * dx ** 2 * dy ** 2 / (2 * self.diffusion * (dx ** 2 + dy ** 2))

        # volume, to convert between counts and concentration
        self.bin_volume = get_bin_volume(self.n_bins, self.bounds, depth)

        # initialize gradient fields
        gradient = initial_parameters.get('gradient', self.defaults['gradient'])
        if gradient:
            gradient_fields = make_gradient(gradient, self.n_bins, self.bounds)
            self.initial_state.update(gradient_fields)

        parameters = {}
        parameters.update(initial_parameters)

        super(DiffusionField, self).__init__(parameters)

    def ports_schema(self):
        local_concentration_schema = {
            molecule: {
                '_default': 0.0,
                '_updater': 'set'}
            for molecule in self.molecule_ids}

        # agents glob schema
        schema = {
            'agents': {
                '*': {
                    'boundary': {
                        'location': {
                            '_default': [0.5 * bound for bound in self.bounds],
                            '_updater': 'set'},
                        'external': local_concentration_schema}}}}

        # fields
        fields_schema = {
            'fields': {
                field: {
                    '_value': self.initial_state.get(field, self.ones_field()),
                    '_updater': 'nonnegative_accumulate',
                    '_emit': True,
                }
                for field in self.molecule_ids
            },
        }
        schema.update(fields_schema)

        # dimensions
        dimensions_schema = {
            'dimensions': {
                'bounds': {
                    '_value': self.parameters['bounds'],
                    '_updater': 'set',
                    '_emit': True,
                },
                'n_bins': {
                    '_value': self.parameters['n_bins'],
                    '_updater': 'set',
                    '_emit': True,
                },
                'depth': {
                    '_value': self.parameters['depth'],
                    '_updater': 'set',
                    '_emit': True,
                }
            },
        }
        schema.update(dimensions_schema)
        return schema

    def next_update(self, timestep, states):
        fields = states['fields']
        agents = states['agents']

        # diffuse field
        delta_fields = self.diffuse(fields, timestep)

        # get each agent's local environment
        local_environments = self.get_local_environments(agents, fields)

        update = {'fields': delta_fields}
        if local_environments:
            update.update({'agents': local_environments})

        return update

    def count_to_concentration(self, count):
        return count_to_concentration(
            count * units.count, self.bin_volume * units.L
        ).to(units.mmol / units.L).magnitude

    def get_bin_site(self, location):
        return get_bin_site(location, self.n_bins, self.bounds)

    def get_single_local_environments(self, specs, fields):
        bin_site = self.get_bin_site(specs['location'])
        local_environment = {}
        for mol_id, field in fields.items():
            local_environment[mol_id] = field[bin_site]
        return local_environment

    def get_local_environments(self, agents, fields):
        local_environments = {}
        if agents:
            for agent_id, specs in agents.items():
                local_environments[agent_id] = {'boundary': {}}
                local_environments[agent_id]['boundary']['external'] = \
                    self.get_single_local_environments(specs['boundary'], fields)
        return local_environments

    def ones_field(self):
        return np.ones((self.n_bins[0], self.n_bins[1]), dtype=np.float64)

    # diffusion functions
    def diffusion_delta(self, field, timestep):
        ''' calculate concentration changes cause by diffusion'''
        field_new = field.copy()
        t = 0.0
        dt = min(timestep, self.diffusion_dt)
        while t < timestep:
            field_new += self.diffusion * dt * convolve(field_new, LAPLACIAN_2D, mode='reflect')
            t += dt

        return field_new - field

    def diffuse(self, fields, timestep):
        delta_fields = {}
        for mol_id, field in fields.items():

            # run diffusion if molecule field is not uniform
            if len(set(field.flatten())) != 1:
                delta = self.diffusion_delta(field, timestep)
            else:
                delta = np.zeros_like(field)
            delta_fields[mol_id] = delta

        return delta_fields


# testing
def get_random_field_config(config={}):
    bounds = config.get('bounds', (20, 20))
    n_bins = config.get('n_bins', (10, 10))
    return {
        'molecules': ['glc'],
        'initial_state': {
            'glc': np.random.rand(n_bins[0], n_bins[1])},
        'n_bins': n_bins,
        'bounds': bounds}

def get_gaussian_config(config={}):
    molecules = config.get('molecules', ['glc'])
    bounds = config.get('bounds', (50, 50))
    n_bins = config.get('n_bins', (20, 20))
    center = config.get('center', [0.5, 0.5])
    deviation = config.get('deviation', 5)
    diffusion = config.get('diffusion', 5e-1)

    return {
        'molecules': molecules,
        'n_bins': n_bins,
        'bounds': bounds,
        'diffusion': diffusion,
        'gradient': {
            'type': 'gaussian',
            'molecules': {
                'glc': {
                    'center': center,
                    'deviation': deviation}}}}

def get_exponential_config(config={}):
    molecules = config.get('molecules', ['glc'])
    bounds = config.get('bounds', (40, 40))
    n_bins = config.get('n_bins', (20, 20))
    center = config.get('center', [1.0, 1.0])
    base = config.get('base', 1 + 2e-4)
    scale = config.get('scale', 0.1)
    diffusion = config.get('diffusion', 1e1)

    return {
        'molecules': molecules,
        'n_bins': n_bins,
        'bounds': bounds,
        'diffusion': diffusion,
        'gradient': {
            'type': 'exponential',
            'molecules': {
                'glc': {
                    'center': center,
                    'base': base,
                    'scale': scale}}}}

def test_diffusion_field(
        config={},
        initial_state={},
        time=10,
):
    diffusion = DiffusionField(config)
    settings = {
        'return_raw_data': True,
        'initial_state': initial_state,
        'total_time': time,
        'timestep': 1}
    return simulate_process(diffusion, settings)

def plot_fields(data, config, out_dir='out', filename='fields'):
    fields = {time: time_data['fields'] for time, time_data in data.items()}
    snapshots_data = {
        'fields': fields,
        'config': config}
    plot_config = {
        'out_dir': out_dir,
        'filename': filename}
    plot_snapshots(snapshots_data, plot_config)


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    parser = argparse.ArgumentParser(description='diffusion_field')
    parser.add_argument('--random', '-r', action='store_true', default=False)
    parser.add_argument('--gaussian', '-g', action='store_true', default=False)
    parser.add_argument('--exponential', '-e', action='store_true', default=False)
    args = parser.parse_args()
    no_args = (len(sys.argv) == 1)

    if args.random or no_args:
        config = get_random_field_config()
        data = test_diffusion_field(
            config=config,
            initial_state={},
            time=60)
        plot_fields(data, config, out_dir, 'random_field')

    if args.gaussian or no_args:
        config = get_gaussian_config()
        data = test_diffusion_field(
            config=config,
            initial_state={},
            time=60)
        plot_fields(data, config, out_dir, 'gaussian_field')

    if args.exponential or no_args:
        config = get_exponential_config()
        data = test_diffusion_field(
            config=config,
            initial_state={},
            time=60)
        plot_fields(data, config, out_dir, 'exponential_field')



import os
import glob
import setuptools
from distutils.core import setup

with open("README.md", 'r') as readme:
    long_description = readme.read()

# to include data in the package, use MANIFEST.in

setup(
    name='vivarium-cell',
    version='0.0.20',
    packages=[
        'vivarium_cell',
        'vivarium_cell.analysis',
        'vivarium_cell.bigg_models',
        'vivarium_cell.composites',
        'vivarium_cell.data',
        'vivarium_cell.data.chromosomes',
        'vivarium_cell.data.flat',
        'vivarium_cell.data.flat.media',
        'vivarium_cell.data.json_files',
        'vivarium_cell.experiments',
        'vivarium_cell.library',
        'vivarium_cell.parameters',
        'vivarium_cell.plots',
        'vivarium_cell.processes',
        'vivarium_cell.reference_data',
        'vivarium_cell.states'
    ],
    author='Eran Agmon, Ryan Spangler',
    author_email='eagmon@stanford.edu, ryan.spangler@gmail.com',
    url='https://github.com/vivarium-collective/vivarium-cell',
    license='MIT',
    entry_points={
        'console_scripts': []},
    description=(
        'A collection of models for simulating cells with Vivarium.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=[
        'vivarium-core>=0.0.19',
        'cobra',
        'Arpeggio',
        'parsimonious',
        'stochastic-arrow',
        'pymunk',
        'alphashape'])



'''
=====================================
Simulate Antibiotic Import and Export
=====================================
'''


from __future__ import absolute_import, division, print_function

import copy
import os

from vivarium.library.units import units
from vivarium.library.dict_utils import deep_merge, deep_merge_check
from vivarium.core.composition import (
    simulate_process_in_experiment,
    plot_simulation_output,
    flatten_timeseries,
    save_timeseries,
    load_timeseries,
    PROCESS_OUT_DIR,
    assert_timeseries_close,
)

from vivarium_cell.processes.convenience_kinetics import ConvenienceKinetics
from vivarium_cell.data import REFERENCE_DATA_DIR


class AntibioticTransport(ConvenienceKinetics):

    name = 'antibiotic_transport'
    defaults = {
        'pump_kcat': 2e-4,
        'pump_km': 0.6,
        'pump_key': 'AcrAB-TolC',
        'antibiotic_key': 'antibiotic',
        'initial_internal_antibiotic': 0,
        'initial_external_antibiotic': 1,
        'initial_pump': 1,
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}
        super_defaults = super(AntibioticTransport, self).defaults
        deep_merge_check(self.defaults, super_defaults)
        self.defaults.update(super_defaults)
        parameters = copy.deepcopy(self.defaults)
        deep_merge(parameters, initial_parameters)

        kinetics_parameters = {
            'reactions': {
                'export': {
                    'stoichiometry': {
                        ('internal', parameters['antibiotic_key']): -1,
                        ('external', parameters['antibiotic_key']): 1,
                    },
                    'is_reversible': False,
                    'catalyzed by': [
                        ('pump_port', parameters['pump_key'])],
                },
            },
            'kinetic_parameters': {
                'export': {
                    ('pump_port', parameters['pump_key']): {
                        'kcat_f': parameters['pump_kcat'],
                        ('internal', parameters['antibiotic_key']):
                            parameters['pump_km'],
                    },
                },
            },
            'initial_state': {
                'fluxes': {
                    'import': 0.0,
                    'export': 0.0,
                },
                'internal': {
                    parameters['antibiotic_key']: parameters[
                        'initial_internal_antibiotic'],
                },
                'external': {
                    parameters['antibiotic_key']: parameters[
                        'initial_external_antibiotic'],
                },
                'pump_port': {
                    parameters['pump_key']: parameters['initial_pump'],
                },
            },
            'port_ids': ['internal', 'external', 'pump_port']
        }

        super(AntibioticTransport, self).__init__(kinetics_parameters)


def run_antibiotic_transport():
    process = AntibioticTransport()
    settings = {
        'total_time': 4000,
        'environment': {
            'volume': 1e-15 * units.L,
        },
    }
    return simulate_process_in_experiment(process, settings)


def test_antibiotic_transport():
    timeseries = run_antibiotic_transport()
    flattened = flatten_timeseries(timeseries)
    reference = load_timeseries(os.path.join(
        REFERENCE_DATA_DIR, AntibioticTransport.name + '.csv'))
    assert_timeseries_close(flattened, reference)


def main():
    out_dir = os.path.join(PROCESS_OUT_DIR, AntibioticTransport.name)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    timeseries = run_antibiotic_transport()
    plot_settings = {}
    plot_simulation_output(timeseries, plot_settings, out_dir)
    save_timeseries(timeseries, out_dir)


if __name__ == '__main__':
    main()



#: List of dictionaries, one for each amino acid. Each dictionary has
#: keys ``name``, ``abbreviation``, and ``symbol``.
amino_acid_records = [
    {'name': 'Alanine', 'abbreviation': 'Ala', 'symbol': 'A'},
    {'name': 'Arginine', 'abbreviation': 'Arg', 'symbol': 'R'},
    {'name': 'Asparagine', 'abbreviation': 'Asn', 'symbol': 'N'},
    {'name': 'Aspartate', 'abbreviation': 'Asp', 'symbol': 'D'},
    {'name': 'Cysteine', 'abbreviation': 'Cys', 'symbol': 'C'},
    {'name': 'Glutamate', 'abbreviation': 'Glu', 'symbol': 'E'},
    {'name': 'Glutamine', 'abbreviation': 'Gln', 'symbol': 'Q'},
    {'name': 'Glycine', 'abbreviation': 'Gly', 'symbol': 'G'},
    {'name': 'Histidine', 'abbreviation': 'His', 'symbol': 'H'},
    {'name': 'Isoleucine', 'abbreviation': 'Ile', 'symbol': 'I'},
    {'name': 'Leucine', 'abbreviation': 'Leu', 'symbol': 'L'},
    {'name': 'Lysine', 'abbreviation': 'Lys', 'symbol': 'K'},
    {'name': 'Methionine', 'abbreviation': 'Met', 'symbol': 'M'},
    {'name': 'Phenylalanine', 'abbreviation': 'Phe', 'symbol': 'F'},
    {'name': 'Proline', 'abbreviation': 'Pro', 'symbol': 'P'},
    {'name': 'Serine', 'abbreviation': 'Ser', 'symbol': 'S'},
    {'name': 'Threonine', 'abbreviation': 'Thr', 'symbol': 'T'},
    {'name': 'Tryptophan', 'abbreviation': 'Trp', 'symbol': 'W'},
    {'name': 'Tyrosine', 'abbreviation': 'Tyr', 'symbol': 'Y'},
    {'name': 'Valine', 'abbreviation': 'Val', 'symbol': 'V'}]

#: Map from amino acid symbol to the amino acid name
amino_acids = {
    record['symbol']: record['name']
    for record in amino_acid_records}



'''
==================
Metabolism Process
==================

This module defines a :term:`process class` for modeling a cell's
metabolic processes with flux balance analysis (FBA). The cobrapy
FBA library is used for solving the problems. This supports metabolic
models from the `BiGG model database <http://bigg.ucsd.edu>`_,
and other configurations that can be passed to :py:class:`Metabolism`
to create models of metabolism.
'''

from __future__ import absolute_import, division, print_function

import os
import argparse
import logging as log

import numpy as np
import matplotlib.pyplot as plt

from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    save_timeseries,
    flatten_timeseries,
    load_timeseries,
    assert_timeseries_close,
    PROCESS_OUT_DIR,
    plot_simulation_output
)
from vivarium.library.make_network import (
    get_reactions,
    make_network,
    save_network
)
from vivarium.library.units import units
from vivarium.library.dict_utils import tuplify_port_dicts

from vivarium_cell.processes.derive_globals import AVOGADRO
from vivarium_cell.library.cobra_fba import CobraFBA
from vivarium_cell.library.regulation_logic import build_rule
from vivarium_cell.data.synonyms import get_synonym
from vivarium_cell.data import REFERENCE_DATA_DIR

# plots
from vivarium_cell.plots.metabolism import (
    plot_exchanges,
    BiGG_energy_carriers,
    energy_synthesis_plot
)


NAME = 'metabolism'


def get_fg_from_counts(counts_dict, mw):
    composition_mass = sum([
        coeff / AVOGADRO * mw.get(mol_id, 0.0) * (units.g / units.mol)
        for mol_id, coeff in counts_dict.items()])  # g
    return composition_mass.to('fg')


def get_minimal_media_iAF1260b(
        scale_concentration=1,
        override_initial={},
):
    config = get_iAF1260b_config()
    metabolism = Metabolism(config)
    molecules = {
        mol_id: conc * scale_concentration
        for mol_id, conc in metabolism.initial_state['external'].items()
    }
    for mol_id, conc in override_initial.items():
        molecules[mol_id] = conc
    return molecules



class Metabolism(Process):
    """A general class that is configured to match specific models

    This metabolism process class models metabolism using flux balance
    analysis (FBA). The FBA problem is defined using the provided
    configuration parameters.

    To see how to configure the process manually, look at the source
    code for :py:func:`test_toy_metabolism`.

    :term:`Ports`:

    * **external**: Holds the state of molecules external to the FBA
      reactions. For a model of a cell's metabolism, this will likely
      hold metabolite concentrations in the extracellular space.
    * **internal**: Holds the state of molecules internal to the FBA.
      For a model of a cell's metabolism, this will probably be the
      cytosolic concentrations.
    * **reactions**: Holds the IDs of the modeled metabolic reactions.
      The linked :term:`store` does not need to be shared with any other
      processes.
    * **fields**: The environmental fields that will be updated with
      cell intake and uptake.
    * **dimensions**: Holds the dimensions of the environment.
    * **flux_bounds**: The bounds on the FBA, which are imposed by the
      availability of metabolites. For example, for the metabolism of a
      cell, the bounds represent the limits of transmembrane transport.
    * **global**: Should be linked to the ``global`` :term:`store`.

    Args:
        initial_parameters (dict): Configures the process with the
            following keys/values:

            * **initial_state** (:py:class:`dict`): the default state,
              with a dict for internal and external, like this:
              ``{'external': external_state, 'internal':
              internal_state}``
            * **stoichiometry** (:py:class:`dict`): a map from reaction
              ID to that reaction's stoichiometry dictionary, e.g.
              ``{reaction_id: stoichiometry_dict}``
            * **objective** (:py:class:`dict`): the stoichiometry dict
              to be optimized
            * **external_molecules** (:py:class:`list`): the external
              molecules
            * **reversible_reactions** (:py:class:`list`)

    """

    name = NAME
    defaults = {
        'constrained_reaction_ids': [],
        'model_path': 'vivarium_cell/bigg_models/iAF1260b.json',
        'default_upper_bound': 0.0,
        'regulation': {},
        'initial_state': {},
        'exchange_threshold': 1e-4,  # concentrations lower than exchange_threshold are considered depleted
        'initial_mass': 1339,  # fg
        'global_deriver_key': 'global_deriver',
        'mass_deriver_key': 'mass_deriver',
        'time_step': 1,
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters == None:
            initial_parameters = {}
        self.nAvogadro = AVOGADRO

        time_step = self.or_default(
            initial_parameters, 'time_step')

        # initialize FBA
        if 'model_path' not in initial_parameters and 'stoichiometry' not in initial_parameters:
            initial_parameters['model_path'] = self.defaults['model_path']
        self.fba = CobraFBA(initial_parameters)
        self.reaction_ids = self.fba.reaction_ids()
        self.exchange_threshold = self.defaults['exchange_threshold']

        # additional FBA options
        self.constrained_reaction_ids = self.or_default(
            initial_parameters, 'constrained_reaction_ids')
        self.default_upper_bound = self.or_default(
            initial_parameters, 'default_upper_bound')

        # make the regulation functions
        regulation_logic = self.or_default(
            initial_parameters, 'regulation')
        self.regulation = {
            reaction: build_rule(logic)
            for reaction, logic in regulation_logic.items()}

        # get internal molecules from fba objective
        self.objective_composition = {}
        for reaction_id, coeff1 in self.fba.objective.items():
            for mol_id, coeff2 in self.fba.stoichiometry[reaction_id].items():
                if mol_id in self.objective_composition:
                    self.objective_composition[mol_id] += coeff1 * coeff2
                else:
                    self.objective_composition[mol_id] = coeff1 * coeff2

        ## Get initial internal state from initial_mass
        initial_metabolite_mass = self.or_default(
            initial_parameters, 'initial_mass')
        mw = self.fba.molecular_weights
        composition = {
            mol_id: (-coeff if coeff < 0 else 0)
            for mol_id, coeff in self.objective_composition.items()}
        composition_mass = get_fg_from_counts(composition, mw)
        scaling_factor = (initial_metabolite_mass / composition_mass).magnitude
        internal_state = {mol_id: int(coeff * scaling_factor)
            for mol_id, coeff in composition.items()}
        self.initial_mass = get_fg_from_counts(internal_state, mw)
        log.info('metabolism initial mass: {}'.format(self.initial_mass))

        ## Get external state from minimal_external fba solution
        external_state = {state_id: 0.0 for state_id in self.fba.external_molecules}
        external_state.update(self.fba.minimal_external)  # optimal minimal media from fba

        # save initial state
        self.initial_state = {
            'external': external_state,
            'internal': internal_state,
            'flux_bounds': {reaction_id: self.default_upper_bound
                            for reaction_id in self.constrained_reaction_ids},
        }

        # parameters
        parameters = {'time_step': time_step}
        parameters.update(initial_parameters)

        self.global_deriver_key = self.or_default(
            initial_parameters, 'global_deriver_key')
        self.mass_deriver_key = self.or_default(
            initial_parameters, 'mass_deriver_key')

        super(Metabolism, self).__init__(parameters)

    def ports_schema(self):
        ports = [
            'internal',
            'external',
            'fields',
            'reactions',
            'flux_bounds',
            'global',
            'dimensions',
        ]

        schema = {port: {} for port in ports}

        # internal
        for state in list(self.objective_composition.keys()):
            schema['internal'][state] = {
                '_value': self.initial_state['internal'].get(state, 0),
                '_divider': 'split',
                '_default': 0.0,
                '_emit': True,
                '_properties': {
                    'mw': self.fba.molecular_weights[state] * units.g / units.mol},
            }

        # external
        for state in self.fba.external_molecules:
            schema['external'][state] = {
                '_default': self.initial_state['external'].get(state, 0.0),
                '_emit': True,
            }

        # fields
        for state in self.fba.external_molecules:
            schema['fields'][state] = {
                '_default': np.zeros((1, 1)),
            }

        # reactions
        for state in self.reaction_ids:
            schema['reactions'][state] = {
                '_default': 0.0,
                '_emit': state in self.constrained_reaction_ids,
                '_updater': 'set',
            }

        # flux_bounds
        for state in self.constrained_reaction_ids:
            schema['flux_bounds'][state] = {
                '_default': self.initial_state['flux_bounds'].get(state, self.default_upper_bound),
                '_emit': True,
            }

        # globals
        schema['global'] = {
            'mass': {
                '_default': self.initial_mass,
                '_emit': True,
            },
            'mmol_to_counts': {
                '_default': 0.0 * units.L / units.mmol,
                '_emit': True,
            },
            'location': {
                '_default': [0.5, 0.5],
            }
        }

        # dimensions
        schema['dimensions'] = {
            'bounds': {
                '_default': [1, 1],
            },
            'n_bins': {
                '_default': [1, 1],
            },
            'depth': {
                '_default': 1,
            },
        }

        return schema

    def derivers(self):
        return {
            self.mass_deriver_key: {
                'deriver': 'mass_deriver',
                'port_mapping': {
                    'global': 'global',
                },
                'config': {
                    'from_path': ('..', '..'),
                },
            },
            self.global_deriver_key: {
                'deriver': 'globals_deriver',
                'port_mapping': {
                    'global': 'global',
                },
                'config': {
                    'initial_mass': self.initial_mass,
                },
            }
        }

    def next_update(self, timestep, states):
        ## get the state
        external_state = states['external']
        constrained_reaction_bounds = states['flux_bounds']  # (units.mmol / units.L / units.s)
        mmol_to_counts = states['global']['mmol_to_counts']

        ## get flux constraints
        # exchange_constraints based on external availability
        exchange_constraints = {mol_id: 0.0
            for mol_id, conc in external_state.items() if conc <= self.exchange_threshold}

        # get state of regulated reactions (True/False)
        flattened_states = tuplify_port_dicts(states)
        regulation_state = {}
        for reaction_id, reg_logic in self.regulation.items():
            regulation_state[reaction_id] = reg_logic(flattened_states)

        ## apply flux constraints
        # first, add exchange constraints
        self.fba.set_exchange_bounds(exchange_constraints)

        # next, add constraints coming from flux_bounds
        # to constrain exchange fluxes, add the suffix 'EX_' to the external molecule ID
        if constrained_reaction_bounds:
            self.fba.constrain_flux(constrained_reaction_bounds)

        # finally, turn reactions on/off based on regulation
        self.fba.regulate_flux(regulation_state)

        ## solve the fba problem
        objective_exchange = self.fba.optimize() * timestep  # (units.mmol / units.L / units.s)
        exchange_reactions = self.fba.read_exchange_reactions()
        exchange_fluxes = self.fba.read_exchange_fluxes()  # (units.mmol / units.L / units.s)
        internal_fluxes = self.fba.read_internal_fluxes()  # (units.mmol / units.L / units.s)

        # timestep dependence on fluxes
        exchange_fluxes.update((mol_id, flux * timestep) for mol_id, flux in exchange_fluxes.items())
        internal_fluxes.update((mol_id, flux * timestep) for mol_id, flux in internal_fluxes.items())

        # update internal counts from objective flux
        # calculate added mass from the objective molecules' molecular weights
        objective_count = (objective_exchange * mmol_to_counts).magnitude
        internal_state_update = {}
        for reaction_id, coeff1 in self.fba.objective.items():
            for mol_id, coeff2 in self.fba.stoichiometry[reaction_id].items():
                if coeff2 < 0:  # pull out molecule if it is USED to make biomass (negative coefficient)
                    added_count = int(-coeff1 * coeff2 * objective_count)
                    internal_state_update[mol_id] = added_count

        # convert exchange fluxes to counts
        field_updates = {
            reaction: {
                '_value': int((flux * mmol_to_counts).magnitude),
                '_updater': {
                    'updater': 'update_field_with_exchange',
                    'port_mapping': {
                        'global': 'global',
                        'dimensions': 'dimensions',
                    },
                },
            }
            for reaction, flux in exchange_fluxes.items()
        }

        all_fluxes = {}
        all_fluxes.update(internal_fluxes)
        all_fluxes.update(exchange_reactions)

        return {
            'fields': field_updates,
            'internal': internal_state_update,
            'reactions': all_fluxes,
        }



# configs
def get_package_path():
    return os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

def get_e_coli_core_config():
    """Get an *E. coli* core metabolism model

    The model is the `e_coli_core model from BiGG
    <http://bigg.ucsd.edu/models/e_coli_core>`_.

    Returns:
        A configuration for the model that can be passed to the
        :py:class:`Metabolism` constructor.
    """
    package_path = get_package_path()
    metabolism_file = os.path.join(package_path, 'bigg_models', 'e_coli_core.json')
    return {'model_path': metabolism_file}

def get_iAF1260b_config():
    """Get the metabolism config for the iAF1260b BiGG model

    The metabolism model is the `iAF1260b model from BiGG
    <http://bigg.ucsd.edu/models/iAF1260b>`_.

    Returns:
        A configuration for the model that can be passed to the
        :py:class:`Metabolism` constructor.
    """
    package_path = get_package_path()
    metabolism_file = os.path.join(package_path, 'bigg_models', 'iAF1260b.json')
    return {'model_path': metabolism_file}

def get_toy_configuration():
    stoichiometry = {
        'R1': {'A': -1, 'ATP': -1, 'B': 1},
        'R2a': {'B': -1, 'ATP': 2, 'NADH': 2, 'C': 1},
        'R2b': {'C': -1, 'ATP': -2, 'NADH': -2, 'B': 1},
        'R3': {'B': -1, 'F': 1},
        'R4': {'C': -1, 'G': 1},
        'R5': {'G': -1, 'C': 0.8, 'NADH': 2},
        'R6': {'C': -1, 'ATP': 2, 'D': 3},
        'R7': {'C': -1, 'NADH': -4, 'E': 3},
        'R8a': {'G': -1, 'ATP': -1, 'NADH': -2, 'H': 1},
        'R8b': {'G': 1, 'ATP': 1, 'NADH': 2, 'H': -1},
        'Rres': {'NADH': -1, 'O2': -1, 'ATP': 1},
        'v_biomass': {'C': -1, 'F': -1, 'H': -1, 'ATP': -10}}

    external_molecules = ['A', 'F', 'D', 'E', 'H', 'O2']

    objective = {'v_biomass': 1.0}

    reversible = ['R6', 'R7', 'Rres']

    default_reaction_bounds = 1000.0

    exchange_bounds = {
        'A': -0.02,
        'D': 0.01,
        'E': 0.01,
        'F': -0.005,
        'H': -0.005,
        'O2': -0.1}

    initial_state = {
        'external': {
            'A': 21.0,
            'F': 5.0,
            'D': 12.0,
            'E': 12.0,
            'H': 5.0,
            'O2': 100.0}}

    # molecular weight units are (units.g / units.mol)
    molecular_weights = {
        'A': 500.0,
        'B': 500.0,
        'C': 500.0,
        'D': 500.0,
        'E': 500.0,
        'F': 50000.0,
        'H': 1.00794,
        'O2': 31.9988,
        'ATP': 507.181,
        'NADH': 664.425}

    config = {
        'stoichiometry': stoichiometry,
        'reversible': reversible,
        'external_molecules': external_molecules,
        'objective': objective,
        'initial_state': initial_state,
        'exchange_bounds': exchange_bounds,
        'default_upper_bound': default_reaction_bounds,
        'molecular_weights': molecular_weights}

    return config


# toy functions
def make_kinetic_rate(mol_id, vmax, km=0.0):
    def rate(state):
        flux = (vmax * state[mol_id]) / (km + state[mol_id])
        return flux
    return rate

def toy_transport():
    # process-like function for transport kinetics, used by simulate_metabolism
    transport_kinetics = {
        "EX_A": make_kinetic_rate("A", -1e-1, 5),  # A import
    }
    return transport_kinetics

# sim functions
def run_sim_save_network(config=get_toy_configuration(), out_dir='out/network'):
    metabolism = Metabolism(config)

    # initialize the process
    stoichiometry = metabolism.fba.stoichiometry
    reaction_ids = list(stoichiometry.keys())
    external_mol_ids = metabolism.fba.external_molecules
    objective = metabolism.fba.objective

    settings = {
        # 'environment_volume': 1e-6,  # L   # TODO -- bring back environment?
        'timestep': 1,
        'total_time': 10}

    timeseries = simulate_process_in_experiment(metabolism, settings)
    reactions = timeseries['reactions']

    # save fluxes as node size
    reaction_fluxes = {}
    for rxn_id in reaction_ids:
        if rxn_id in reactions:
            flux = abs(np.mean(reactions[rxn_id][1:]))
            reaction_fluxes[rxn_id] = np.log(1000 * flux + 1.1)
        else:
            reaction_fluxes[rxn_id] = 1

    # define node type
    node_types = {rxn_id: 'reaction' for rxn_id in reaction_ids}
    node_types.update({mol_id: 'external_mol' for mol_id in external_mol_ids})
    node_types.update({rxn_id: 'objective' for rxn_id in objective.keys()})
    info = {
        'node_types': node_types,
        'reaction_fluxes': reaction_fluxes}

    nodes, edges = make_network(stoichiometry, info)
    save_network(nodes, edges, out_dir)

def run_metabolism(metabolism, settings=None):
    if not settings:
        settings = {
            'total_time': 10}
    return simulate_process_in_experiment(metabolism, settings)

# tests
def test_toy_metabolism():
    regulation_logic = {
        'R4': 'if (external, O2) > 0.1 and not (external, F) < 0.1'}

    toy_config = get_toy_configuration()
    transport = toy_transport()

    toy_config['constrained_reaction_ids'] = list(transport.keys())
    toy_config['regulation'] = regulation_logic
    toy_metabolism = Metabolism(toy_config)

    # TODO -- add molecular weights!

    # simulate toy model
    timeline = [
        (5, {('external', 'A'): 1}),
        (10, {('external', 'F'): 0}),
        (15, {})]

    settings = {
        'environment': {
            'volume': 1e-8 * units.L,
        },
        'timestep': 1.0,
        'timeline': {
            'timeline': timeline}}
    return simulate_process_in_experiment(toy_metabolism, settings)

def test_BiGG_metabolism(config=get_iAF1260b_config(), settings={}):
    metabolism = Metabolism(config)
    run_metabolism(metabolism, settings)

reference_sim_settings = {
    'environment': {
        'volume': 1e-5 * units.L,
    },
    'timestep': 1,
    'total_time': 10}

def test_metabolism_similar_to_reference():
    config = get_iAF1260b_config()
    metabolism = Metabolism(config)
    timeseries = run_metabolism(metabolism, reference_sim_settings)

    reference = load_timeseries(
        os.path.join(REFERENCE_DATA_DIR, NAME + '.csv'))
    assert_timeseries_close(timeseries, reference)


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    parser = argparse.ArgumentParser(description='metabolism process')
    parser.add_argument('--bigg', '-b', action='store_true', default=False,)
    args = parser.parse_args()

    if args.bigg:
        # configure BiGG metabolism
        config = get_iAF1260b_config()
        metabolism = Metabolism(config)
        external_concentrations = metabolism.initial_state['external']

        # simulation settings
        sim_settings = {
            'environment': {
                'volume': 1e-5 * units.L,
                'concentrations': external_concentrations,
            },
            'total_time': 2520,  # 2520 sec (42 min) is the expected doubling time in minimal media
        }

        # run simulation
        timeseries = simulate_process_in_experiment(metabolism, sim_settings)

        save_timeseries(timeseries, out_dir)
        volume_ts = timeseries['global']['volume']
        mass_ts = timeseries['global']['mass']
        print('volume growth: {}'.format(volume_ts[-1] / volume_ts[0]))
        print('mass growth: {}'.format(mass_ts[-1] / mass_ts[0]))

        # plot settings
        plot_settings = {
            'max_rows': 30,
            'remove_zeros': True,
            'skip_ports': ['exchange', 'reactions']}

        # make plots from simulation output
        plot_simulation_output(timeseries, plot_settings, out_dir, 'BiGG_simulation')
        plot_exchanges(timeseries, sim_settings, out_dir)

        # # make plot of energy reactions
        # stoichiometry = metabolism.fba.stoichiometry
        # energy_carriers = [get_synonym(mol_id) for mol_id in BiGG_energy_carriers]
        # energy_reactions = get_reactions(stoichiometry, energy_carriers)
        # energy_plot_settings = {'reactions': energy_reactions}
        # energy_synthesis_plot(timeseries, energy_plot_settings, out_dir)

        # make a gephi network
        run_sim_save_network(get_iAF1260b_config(), out_dir)

    else:
        timeseries = test_toy_metabolism()
        plot_settings = {}
        plot_simulation_output(timeseries, plot_settings, out_dir, 'toy_metabolism')



from __future__ import absolute_import, division, print_function

import os
import math
import random
import itertools

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
from matplotlib.lines import Line2D
from matplotlib.colors import hsv_to_rgb
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.collections import LineCollection
import numpy as np

from vivarium.library.dict_utils import get_value_from_path


DEFAULT_BOUNDS = [10, 10]

# constants
PI = math.pi

# colors for phylogeny initial agents
HUES = [hue/360 for hue in np.linspace(0,360,30)]
DEFAULT_HUE = HUES[0]
DEFAULT_SV = [100.0/100.0, 70.0/100.0]
BASELINE_TAG_COLOR = [220/360, 1.0, 0.2]  # HSV
FLOURESCENT_SV = [0.75, 1.0]  # SV for fluorescent colors

def check_plt_backend():
    # reset matplotlib backend for non-interactive plotting
    plt.close('all')
    if plt.get_backend() == 'TkAgg':
        matplotlib.use('Agg')

class LineWidthData(Line2D):
    def __init__(self, *args, **kwargs):
        _lw_data = kwargs.pop('linewidth', 1)
        super().__init__(*args, **kwargs)
        self._lw_data = _lw_data

    def _get_lw(self):
        if self.axes is not None:
            ppd = 72./self.axes.figure.dpi
            trans = self.axes.transData.transform
            return ((trans((1, self._lw_data))-trans((0, 0)))*ppd)[1]
        else:
            return 1

    def _set_lw(self, lw):
        self._lw_data = lw

    _linewidth = property(_get_lw, _set_lw)

def plot_agent(ax, data, color, agent_shape):
    # location, orientation, length
    x_center = data['boundary']['location'][0]
    y_center = data['boundary']['location'][1]
    theta = data['boundary']['angle'] / PI * 180 + 90 # rotate 90 degrees to match field
    length = data['boundary']['length']
    width = data['boundary']['width']

    # get color, convert to rgb
    rgb = hsv_to_rgb(color)

    if agent_shape == 'rectangle':
        # get bottom left position
        x_offset = (width / 2)
        y_offset = (length / 2)
        theta_rad = math.radians(theta)
        dx = x_offset * math.cos(theta_rad) - y_offset * math.sin(theta_rad)
        dy = x_offset * math.sin(theta_rad) + y_offset * math.cos(theta_rad)

        x = x_center - dx
        y = y_center - dy

        # Create a rectangle
        shape = patches.Rectangle(
            (x, y), width, length,
            angle=theta,
            linewidth=2,
            edgecolor='w',
            facecolor=rgb
        )
        ax.add_patch(shape)

    elif agent_shape == 'segment':
        membrane_width = 0.1
        membrane_color = [1, 1, 1]
        radius = width / 2

        # get the two ends
        length_offset = (length / 2) - radius
        theta_rad = math.radians(theta)
        dx = - length_offset * math.sin(theta_rad)
        dy = length_offset * math.cos(theta_rad)

        x1 = x_center - dx
        y1 = y_center - dy
        x2 = x_center + dx
        y2 = y_center + dy

        # segment plot
        membrane = LineWidthData(
            [x1, x2], [y1, y2],
            color=membrane_color,
            linewidth=width,
            solid_capstyle='round')
        line = LineWidthData(
            [x1, x2], [y1, y2],
            color=rgb,
            linewidth=width-membrane_width,
            solid_capstyle='round')
        ax.add_line(membrane)
        ax.add_line(line)


def plot_agents(
    ax, agents, agent_colors={}, agent_shape='segment', dead_color=None
):
    '''
    - ax: the axis for plot
    - agents: a dict with {agent_id: agent_data} and
        agent_data a dict with keys location, angle, length, width
    - agent_colors: dict with {agent_id: hsv color}
    - dead_color: List of 3 floats that define HSV color to use for dead
      cells. Dead cells only get treated differently if this is set.
    '''
    for agent_id, agent_data in agents.items():
        color = agent_colors.get(agent_id, [DEFAULT_HUE]+DEFAULT_SV)
        if dead_color and 'boundary' in agent_data and 'dead' in agent_data['boundary']:
            if agent_data['boundary']['dead']:
                color = dead_color
        plot_agent(ax, agent_data, color, agent_shape)

def mutate_color(baseline_hsv):
    mutation = 0.1
    new_hsv = [
        (n + np.random.uniform(-mutation, mutation))
        for n in baseline_hsv]
    # wrap hue around
    new_hsv[0] = new_hsv[0] % 1
    # reflect saturation and value
    if new_hsv[1] > 1:
        new_hsv[1] = 2 - new_hsv[1]
    if new_hsv[2] > 1:
        new_hsv[2] = 2 - new_hsv[2]
    return new_hsv

def color_phylogeny(ancestor_id, phylogeny, baseline_hsv, phylogeny_colors={}):
    """
    get colors for all descendants of the ancestor
    through recursive calls to each generation
    """
    phylogeny_colors.update({ancestor_id: baseline_hsv})
    daughter_ids = phylogeny.get(ancestor_id)
    if daughter_ids:
        for daughter_id in daughter_ids:
            daughter_color = mutate_color(baseline_hsv)
            color_phylogeny(daughter_id, phylogeny, daughter_color)
    return phylogeny_colors

def get_phylogeny_colors_from_names(agent_ids):
    '''Get agent colors using phlogeny saved in agent_ids
    This assumes the names use daughter_phylogeny_id() from meta_division
    '''

    # make phylogeny with {mother_id: [daughter_1_id, daughter_2_id]}
    phylogeny = {agent_id: [] for agent_id in agent_ids}
    for agent1, agent2 in itertools.combinations(agent_ids, 2):
        if agent1 == agent2[0:-1]:
            phylogeny[agent1].append(agent2)
        elif agent2 == agent1[0:-1]:
            phylogeny[agent2].append(agent1)

    # get initial ancestors
    daughters = list(phylogeny.values())
    daughters = set([item for sublist in daughters for item in sublist])
    mothers = set(list(phylogeny.keys()))
    ancestors = list(mothers - daughters)

    # agent colors based on phylogeny
    agent_colors = {agent_id: [] for agent_id in agent_ids}
    for agent_id in ancestors:
        hue = random.choice(HUES)  # select random initial hue
        initial_color = [hue] + DEFAULT_SV
        agent_colors.update(color_phylogeny(agent_id, phylogeny, initial_color))

    return agent_colors


def plot_snapshots(data, plot_config):
    '''Plot snapshots of the simulation over time

    The snapshots depict the agents and environmental molecule
    concentrations.

    Arguments:
        data (dict): A dictionary with the following keys:

            * **agents** (:py:class:`dict`): A mapping from times to
              dictionaries of agent data at that timepoint. Agent data
              dictionaries should have the same form as the hierarchy
              tree rooted at ``agents``.
            * **fields** (:py:class:`dict`): A mapping from times to
              dictionaries of environmental field data at that
              timepoint.  Field data dictionaries should have the same
              form as the hierarchy tree rooted at ``fields``.
            * **config** (:py:class:`dict`): The environmental
              configuration dictionary  with the following keys:

                * **bounds** (:py:class:`tuple`): The dimensions of the
                  environment.

        plot_config (dict): Accepts the following configuration options.
            Any options with a default is optional.

            * **n_snapshots** (:py:class:`int`): Number of snapshots to
              show per row (i.e. for each molecule). Defaults to 6.
            * **agent_shape** (:py:class:`str`): the shape of the agents.
              select from **rectangle**, **segment**
            * **phylogeny_names** (:py:class:`bool`): This selects agent
              colors based on phylogenies seved in their names using
              meta_division.py daughter_phylogeny_id()
            * **out_dir** (:py:class:`str`): Output directory, which is
              ``out`` by default.
            * **filename** (:py:class:`str`): Base name of output file.
              ``snapshots`` by default.
            * **skip_fields** (:py:class:`Iterable`): Keys of fields to
              exclude from the plot. This takes priority over
              ``include_fields``.
            * **include_fields** (:py:class:`Iterable`): Keys of fields
              to plot.
            * **field_label_size** (:py:class:`float`): Font size of the
              field label.
            * **dead_color** (:py:class:`list` of 3 :py:class:`float`s):
              Color for dead cells in HSV. Defaults to [0, 0, 0], which
              is black.
            * **default_font_size** (:py:class:`float`): Font size for
              titles and axis labels.
    '''
    check_plt_backend()

    n_snapshots = plot_config.get('n_snapshots', 6)
    out_dir = plot_config.get('out_dir', 'out')
    filename = plot_config.get('filename', 'snapshots')
    agent_shape = plot_config.get('agent_shape', 'segment')
    phylogeny_names = plot_config.get('phylogeny_names', True)
    skip_fields = plot_config.get('skip_fields', [])
    include_fields = plot_config.get('include_fields', None)
    field_label_size = plot_config.get('field_label_size', 20)
    default_font_size = plot_config.get('default_font_size', 36)
    dead_color = plot_config.get('dead_color', [0, 0, 0])

    # get data
    agents = data.get('agents', {})
    fields = data.get('fields', {})
    config = data.get('config', {})
    bounds = config.get('bounds', DEFAULT_BOUNDS)
    edge_length_x = bounds[0]
    edge_length_y = bounds[1]

    # time steps that will be used
    if agents and fields:
        assert set(list(agents.keys())) == set(list(fields.keys())), 'agent and field times are different'
        time_vec = list(agents.keys())
    elif agents:
        time_vec = list(agents.keys())
    elif fields:
        time_vec = list(fields.keys())
    else:
        raise Exception('No agents or field data')

    time_indices = np.round(np.linspace(0, len(time_vec) - 1, n_snapshots)).astype(int)
    snapshot_times = [time_vec[i] for i in time_indices]

    # get fields id and range
    if fields:
        if include_fields is None:
            field_ids = set(fields[time_vec[0]].keys())
        else:
            field_ids = set(include_fields)
        field_ids -= set(skip_fields)
        field_range = {}
        for field_id in field_ids:
            field_min = min([min(min(field_data[field_id])) for t, field_data in fields.items()])
            field_max = max([max(max(field_data[field_id])) for t, field_data in fields.items()])
            field_range[field_id] = [field_min, field_max]

    # get agent ids
    agent_ids = set()
    if agents:
        for time, time_data in agents.items():
            current_agents = list(time_data.keys())
            agent_ids.update(current_agents)
        agent_ids = list(agent_ids)

        # set agent colors
        if phylogeny_names:
            agent_colors = get_phylogeny_colors_from_names(agent_ids)
        else:
            agent_colors = {}
            for agent_id in agent_ids:
                hue = random.choice(HUES)  # select random initial hue
                color = [hue] + DEFAULT_SV
                agent_colors[agent_id] = color

    # make the figure
    n_rows = max(len(field_ids), 1)
    n_cols = n_snapshots + 1  # one column for the colorbar
    figsize = (12 * n_cols, 12 * n_rows)
    max_dpi = min([2**16 // dim for dim in figsize]) - 1
    fig = plt.figure(figsize=figsize, dpi=min(max_dpi, 100))
    grid = plt.GridSpec(n_rows, n_cols, wspace=0.2, hspace=0.2)
    original_fontsize = plt.rcParams['font.size']
    plt.rcParams.update({'font.size': default_font_size})

    # plot snapshot data in each subsequent column
    for col_idx, (time_idx, time) in enumerate(zip(time_indices, snapshot_times)):
        if field_ids:
            for row_idx, field_id in enumerate(field_ids):

                ax = init_axes(
                    fig, edge_length_x, edge_length_y, grid, row_idx,
                    col_idx, time, field_id, field_label_size,
                )
                ax.tick_params(
                    axis='both', which='both', bottom=False, top=False,
                    left=False, right=False,
                )

                # transpose field to align with agents
                field = np.transpose(np.array(fields[time][field_id])).tolist()
                vmin, vmax = field_range[field_id]
                im = plt.imshow(field,
                                origin='lower',
                                extent=[0, edge_length_x, 0, edge_length_y],
                                vmin=vmin,
                                vmax=vmax,
                                cmap='BuPu')
                if agents:
                    agents_now = agents[time]
                    plot_agents(
                        ax, agents_now, agent_colors, agent_shape, dead_color)

                # colorbar in new column after final snapshot
                if col_idx == n_snapshots - 1:
                    cbar_col = col_idx + 1
                    ax = fig.add_subplot(grid[row_idx, cbar_col])
                    if row_idx == 0:
                        ax.set_title('Concentration (mmol/L)', y=1.08)
                    ax.axis('off')
                    if vmin == vmax:
                        continue
                    divider = make_axes_locatable(ax)
                    cax = divider.append_axes("left", size="5%", pad=0.0)
                    fig.colorbar(im, cax=cax, format='%.6f')
                    ax.axis('off')
        else:
            row_idx = 0
            ax = init_axes(
                fig, bounds[0], bounds[1], grid, row_idx, col_idx,
                time, ""
            )

            if agents:
                agents_now = agents[time]
                plot_agents(ax, agents_now, agent_colors, agent_shape, dead_color)

    fig_path = os.path.join(out_dir, filename)
    fig.subplots_adjust(wspace=0.7, hspace=0.1)
    fig.savefig(fig_path, bbox_inches='tight')
    plt.close(fig)
    plt.rcParams.update({'font.size': original_fontsize})

def get_fluorescent_color(baseline_hsv, tag_color, intensity):
    # move color towards bright fluoresence color when intensity = 1
    new_hsv = baseline_hsv[:]
    distance = [a - b for a, b in zip(tag_color, new_hsv)]

    # if hue distance > 180 degrees, go around in the other direction
    if distance[0] > 0.5:
        distance[0] = 1 - distance[0]
    elif distance[0] < -0.5:
        distance[0] = 1 + distance[0]

    new_hsv = [a + intensity * b for a, b in zip(new_hsv, distance)]
    new_hsv[0] = new_hsv[0] % 1

    return new_hsv

def plot_tags(data, plot_config):
    '''Plot snapshots of the simulation over time

    The snapshots depict the agents and the levels of tagged molecules
    in each agent by agent color intensity.

    Arguments:
        data (dict): A dictionary with the following keys:

            * **agents** (:py:class:`dict`): A mapping from times to
              dictionaries of agent data at that timepoint. Agent data
              dictionaries should have the same form as the hierarchy
              tree rooted at ``agents``.
            * **config** (:py:class:`dict`): The environmental
              configuration dictionary  with the following keys:

                * **bounds** (:py:class:`tuple`): The dimensions of the
                  environment.

        plot_config (dict): Accepts the following configuration options.
            Any options with a default is optional.

            * **n_snapshots** (:py:class:`int`): Number of snapshots to
              show per row (i.e. for each molecule). Defaults to 6.
            * **out_dir** (:py:class:`str`): Output directory, which is
              ``out`` by default.
            * **filename** (:py:class:`str`): Base name of output file.
              ``tags`` by default.
            * **tagged_molecules** (:py:class:`typing.Iterable`): The
              tagged molecules whose concentrations will be indicated by
              agent color. Each molecule should be specified as a
              :py:class:`tuple` of the path in the agent compartment
              to where the molecule's count can be found, with the last
              value being the molecule's count variable.
            * **convert_to_concs** (:py:class:`bool`): if True, convert counts
              to concentrations.
            * **background_color** (:py:class:`str`): use matplotlib colors,
              ``black`` by default
            * **tag_label_size** (:py:class:`float`): The font size for
              the tag name label
            * **default_font_size** (:py:class:`float`): Font size for
              titles and axis labels.
    '''
    check_plt_backend()

    n_snapshots = plot_config.get('n_snapshots', 6)
    out_dir = plot_config.get('out_dir', 'out')
    filename = plot_config.get('filename', 'tags')
    agent_shape = plot_config.get('agent_shape', 'segment')
    background_color = plot_config.get('background_color', 'black')
    tagged_molecules = plot_config['tagged_molecules']
    tag_path_name_map = plot_config.get('tag_path_name_map', {})
    tag_label_size = plot_config.get('tag_label_size', 20)
    default_font_size = plot_config.get('default_font_size', 36)
    convert_to_concs = plot_config.get('convert_to_concs', True)

    if tagged_molecules == []:
        raise ValueError('At least one molecule must be tagged.')

    # get data
    agents = data['agents']
    config = data.get('config', {})
    bounds = config['bounds']
    edge_length_x, edge_length_y = bounds

    # time steps that will be used
    time_vec = list(agents.keys())
    time_indices = np.round(
        np.linspace(0, len(time_vec) - 1, n_snapshots)
    ).astype(int)
    snapshot_times = [time_vec[i] for i in time_indices]

    # get tag ids and range
    tag_ranges = {}
    tag_colors = {}

    for time, time_data in agents.items():
        for agent_id, agent_data in time_data.items():
            volume = agent_data.get('boundary', {}).get('volume', 0)
            for tag_id in tagged_molecules:
                level = get_value_from_path(agent_data, tag_id)
                if convert_to_concs:
                    level = level / volume if volume else 0
                if tag_id in tag_ranges:
                    tag_ranges[tag_id] = [
                        min(tag_ranges[tag_id][0], level),
                        max(tag_ranges[tag_id][1], level)]
                else:
                    # add new tag
                    tag_ranges[tag_id] = [level, level]

                    # select random initial hue
                    hue = random.choice(HUES)
                    tag_color = [hue] + FLOURESCENT_SV
                    tag_colors[tag_id] = tag_color

    # make the figure
    n_rows = len(tagged_molecules)
    n_cols = n_snapshots + 1  # one column for the colorbar
    figsize = (12 * n_cols, 12 * n_rows)
    max_dpi = min([2**16 // dim for dim in figsize]) - 1
    fig = plt.figure(figsize=figsize, dpi=min(max_dpi, 100))
    grid = plt.GridSpec(n_rows, n_cols, wspace=0.2, hspace=0.2)
    original_fontsize = plt.rcParams['font.size']
    plt.rcParams.update({'font.size': default_font_size})

    # plot tags
    for row_idx, tag_id in enumerate(tag_ranges.keys()):
        used_agent_colors = []
        concentrations = []
        for col_idx, (time_idx, time) in enumerate(
            zip(time_indices, snapshot_times)
        ):
            tag_name = tag_path_name_map.get(tag_id, tag_id)
            ax = init_axes(
                fig, edge_length_x, edge_length_y, grid,
                row_idx, col_idx, time, tag_name, tag_label_size,
            )
            ax.tick_params(
                axis='both', which='both', bottom=False, top=False,
                left=False, right=False,
            )
            ax.set_facecolor(background_color)

            # update agent colors based on tag_level
            min_tag, max_tag = tag_ranges[tag_id]
            agent_tag_colors = {}
            for agent_id, agent_data in agents[time].items():
                agent_color = BASELINE_TAG_COLOR

                # get current tag concentration, and determine color
                level = get_value_from_path(agent_data, tag_id)
                if convert_to_concs:
                    volume = agent_data.get('boundary', {}).get('volume', 0)
                    level = level / volume if volume else 0
                if min_tag != max_tag:
                    concentrations.append(level)
                    intensity = max((level - min_tag), 0)
                    intensity = min(intensity / (max_tag - min_tag), 1)
                    tag_color = tag_colors[tag_id]
                    agent_color = get_fluorescent_color(
                        BASELINE_TAG_COLOR, tag_color, intensity)
                    agent_rgb = matplotlib.colors.hsv_to_rgb(agent_color)
                    used_agent_colors.append(agent_rgb)

                agent_tag_colors[agent_id] = agent_color

            plot_agents(ax, agents[time], agent_tag_colors, agent_shape)

            # colorbar in new column after final snapshot
            if col_idx == n_snapshots - 1:
                cbar_col = col_idx + 1
                ax = fig.add_subplot(grid[row_idx, cbar_col])
                if row_idx == 0:
                    if convert_to_concs:
                        ax.set_title('Concentration (counts/fL)', y=1.08)
                ax.axis('off')
                if min_tag == max_tag:
                    continue
                divider = make_axes_locatable(ax)
                cax = divider.append_axes("left", size="5%", pad=0.0)
                norm = matplotlib.colors.Normalize()
                # Sort colors and concentrations by concentration
                sorted_idx = np.argsort(concentrations)
                norm.autoscale(used_agent_colors)
                cmap = matplotlib.colors.ListedColormap(
                    np.array(used_agent_colors)[sorted_idx])
                mappable = matplotlib.cm.ScalarMappable(norm, cmap)
                mappable.set_array(np.array(concentrations)[sorted_idx])
                mappable.set_clim(min_tag, max_tag)
                fig.colorbar(mappable, cax=cax, format='%.6f')

    fig_path = os.path.join(out_dir, filename)
    fig.subplots_adjust(wspace=0.7, hspace=0.1)
    fig.savefig(fig_path, bbox_inches='tight')
    plt.close(fig)
    plt.rcParams.update({'font.size': original_fontsize})

def initialize_spatial_figure(bounds, fontsize=18):

    x_length = bounds[0]
    y_length = bounds[1]

    # set up figure
    n_ticks = 4
    plot_buffer = 0.02
    buffer = plot_buffer * min(bounds)
    min_edge = min(x_length, y_length)
    x_scale = x_length/min_edge
    y_scale = y_length/min_edge

    # make the figure
    fig = plt.figure(figsize=(8*x_scale, 8*y_scale))
    plt.rcParams.update({'font.size': fontsize, "font.family": "Times New Roman"})

    plt.xlim((0-buffer, x_length+buffer))
    plt.ylim((0-buffer, y_length+buffer))
    plt.xlabel(u'\u03bcm')
    plt.ylabel(u'\u03bcm')

    # specify the number of ticks for each edge
    [x_bins, y_bins] = [int(n_ticks * edge / min_edge) for edge in [x_length, y_length]]
    plt.locator_params(axis='y', nbins=y_bins)
    plt.locator_params(axis='x', nbins=x_bins)
    ax = plt.gca()

    return fig, ax

def get_agent_trajectories(agents, times):
    trajectories = {}
    for agent_id, series in agents.items():
        time_indices = series['boundary']['location']['time_index']
        series_times = [times[time_index] for time_index in time_indices]

        positions = series['boundary']['location']['value']
        angles = series['boundary']['angle']['value']
        series_values = [[x, y, theta] for ((x, y), theta) in zip(positions, angles)]

        trajectories[agent_id] = {
            'time': series_times,
            'value': series_values,
        }
    return trajectories

def get_agent_type_colors(agent_ids):
    """ get colors for each agent id by agent type
    Assumes that agents of the same type share the beginning
    of their name, followed by '_x' with x as a single number
    TODO -- make this more general for more digits and other comparisons"""
    agent_type_colors = {}
    agent_types = {}
    for agent1, agent2 in itertools.combinations(agent_ids, 2):
        if agent1[0:-2] == agent2[0:-2]:
            agent_type = agent1[0:-2]
            if agent_type not in agent_type_colors:
                color = plt.rcParams['axes.prop_cycle'].by_key()['color'][len(agent_type_colors)]
                agent_type_colors[agent_type] = color
            else:
                color = agent_type_colors[agent_type]
            agent_types[agent1] = agent_type
            agent_types[agent2] = agent_type
    for agent in agent_ids:
        if agent not in agent_types:
            color = plt.rcParams['axes.prop_cycle'].by_key()['color'][len(agent_type_colors)]
            agent_type_colors[agent] = color
            agent_types[agent] = agent

    return agent_types, agent_type_colors

def plot_agent_trajectory(agent_timeseries, config, out_dir='out', filename='trajectory'):
    check_plt_backend()

    # trajectory plot settings
    legend_fontsize = 18
    markersize = 25

    bounds = config.get('bounds', DEFAULT_BOUNDS)
    field = config.get('field')
    rotate_90 = config.get('rotate_90', False)

    # get agents
    times = np.array(agent_timeseries['time'])
    agents = agent_timeseries['agents']
    agent_types, agent_type_colors = get_agent_type_colors(list(agents.keys()))

    if rotate_90:
        field = rotate_field_90(field)
        for agent_id, series in agents.items():
            agents[agent_id] = rotate_agent_series_90(series, bounds)
        bounds = rotate_bounds_90(bounds)

    # get each agent's trajectory
    trajectories = get_agent_trajectories(agents, times)

    # initialize a spatial figure
    fig, ax = initialize_spatial_figure(bounds, legend_fontsize)

    # move x axis to top
    ax.tick_params(labelbottom=False,labeltop=True,bottom=False,top=True)
    ax.xaxis.set_label_coords(0.5, 1.12)

    if field is not None:
        field = np.transpose(field)
        shape = field.shape
        im = plt.imshow(field,
                        origin='lower',
                        extent=[0, shape[1], 0, shape[0]],
                        cmap='Greys')
        # colorbar for field concentrations
        cbar = plt.colorbar(im, pad=0.02, aspect=50, shrink=0.7)
        cbar.set_label('concentration', rotation=270, labelpad=20)

    for agent_id, trajectory_data in trajectories.items():
        agent_trajectory = trajectory_data['value']

        # convert trajectory to 2D array
        locations_array = np.array(agent_trajectory)
        x_coord = locations_array[:, 0]
        y_coord = locations_array[:, 1]

        # get agent type and color
        agent_type = agent_types[agent_id]
        agent_color = agent_type_colors[agent_type]

        # plot line
        ax.plot(x_coord, y_coord, linewidth=2, color=agent_color, label=agent_type)
        ax.plot(x_coord[0], y_coord[0],
                 color=(0.0, 0.8, 0.0), marker='.', markersize=markersize)  # starting point
        ax.plot(x_coord[-1], y_coord[-1],
                 color='r', marker='.', markersize=markersize)  # ending point

    # create legend for agent types
    agent_labels = [
        mlines.Line2D([], [], color=agent_color, linewidth=2, label=agent_type)
        for agent_type, agent_color in agent_type_colors.items()]
    agent_legend = plt.legend(
        title='agent type', handles=agent_labels, loc='upper center',
        bbox_to_anchor=(0.3, 0.0), ncol=2, prop={'size': legend_fontsize})
    ax.add_artist(agent_legend)

    # create a legend for start/end markers
    start = mlines.Line2D([], [],
            color=(0.0, 0.8, 0.0), marker='.', markersize=markersize, linestyle='None', label='start')
    end = mlines.Line2D([], [],
            color='r', marker='.', markersize=markersize, linestyle='None', label='end')
    marker_legend = plt.legend(
        title='trajectory', handles=[start, end], loc='upper center',
        bbox_to_anchor=(0.7, 0.0), ncol=2, prop={'size': legend_fontsize})
    ax.add_artist(marker_legend)

    fig_path = os.path.join(out_dir, filename)
    plt.subplots_adjust(wspace=0.7, hspace=0.1)
    plt.savefig(fig_path, bbox_inches='tight')
    plt.close(fig)


def rotate_bounds_90(bounds):
    return [bounds[1], bounds[0]]

def rotate_field_90(field):
    return np.rot90(field, 3)  # rotate 3 times for 270

def rotate_agent_series_90(series, bounds):
    location_series = series['boundary']['location']
    angle_series = series['boundary']['angle']

    if isinstance(location_series, dict):
        # this ran with time_indexed_timeseries_from_data
        series['boundary']['location']['value'] = [[y, bounds[0] - x] for [x, y] in location_series['value']]
        series['boundary']['angle']['value'] = [theta + PI / 2 for theta in angle_series['value']]
    else:
        series['boundary']['location'] = [[y, bounds[0] - x] for [x, y] in location_series]
        series['boundary']['angle'] = [theta + PI / 2 for theta in angle_series]
    return series

def plot_temporal_trajectory(agent_timeseries, config, out_dir='out', filename='temporal'):
    check_plt_backend()

    bounds = config.get('bounds', DEFAULT_BOUNDS)
    field = config.get('field')
    rotate_90 = config.get('rotate_90', False)

    # get agents
    times = np.array(agent_timeseries['time'])
    agents = agent_timeseries['agents']

    if rotate_90:
        field = rotate_field_90(field)
        for agent_id, series in agents.items():
            agents[agent_id] = rotate_agent_series_90(series, bounds)
        bounds = rotate_bounds_90(bounds)

    # get each agent's trajectory
    trajectories = get_agent_trajectories(agents, times)

    # initialize a spatial figure
    fig, ax = initialize_spatial_figure(bounds)

    if field is not None:
        field = np.transpose(field)
        shape = field.shape
        im = plt.imshow(field,
                        origin='lower',
                        extent=[0, shape[1], 0, shape[0]],
                        cmap='Greys'
                        )

    for agent_id, trajectory_data in trajectories.items():
        agent_trajectory = trajectory_data['value']

        # convert trajectory to 2D array
        locations_array = np.array(agent_trajectory)
        x_coord = locations_array[:, 0]
        y_coord = locations_array[:, 1]

        # make multi-colored trajectory
        points = np.array([x_coord, y_coord]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        lc = LineCollection(segments, cmap=plt.get_cmap('cool'))
        lc.set_array(times)
        lc.set_linewidth(6)

        # plot line
        line = plt.gca().add_collection(lc)

    # color bar
    cbar = plt.colorbar(line, ticks=[times[0], times[-1]], aspect=90, shrink=0.4)
    cbar.set_label('time (s)', rotation=270)

    fig_path = os.path.join(out_dir, filename)
    plt.subplots_adjust(wspace=0.7, hspace=0.1)
    plt.savefig(fig_path, bbox_inches='tight')
    plt.close(fig)

def init_axes(
    fig, edge_length_x, edge_length_y, grid, row_idx, col_idx, time,
    molecule, ylabel_size=20
):
    ax = fig.add_subplot(grid[row_idx, col_idx])
    if row_idx == 0:
        plot_title = 'time: {:.4f} s'.format(float(time))
        plt.title(plot_title, y=1.08)
    if col_idx == 0:
        ax.set_ylabel(
            molecule, fontsize=ylabel_size, rotation='horizontal',
            horizontalalignment='right',
        )
    ax.set(xlim=[0, edge_length_x], ylim=[0, edge_length_y], aspect=1)
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    return ax



from __future__ import absolute_import, division, print_function

import numpy as np
from scipy import constants

from vivarium.library.dict_utils import deep_merge
from vivarium.library.units import units, remove_units, Quantity
from vivarium.core.process import Process


AVOGADRO = constants.N_A * 1 / units.mol


class CellEnvironmentDiffusionFicks(Process):

    name = "cell_environment_diffusion_ficks"
    defaults = {
        'molecules_to_diffuse': [],
        'default_state': {
            'global': {
                'volume': 1.2 * units.fL,
            }
        },
        'default_default': 0,
        # This is an order-of-magnitude estimate from Nagano, K., &
        # Nikaido, H. (2009). Kinetic behavior of the major multidrug
        # efflux pump AcrB of Escherichia coli. Proceedings of the
        # National Academy of Sciences, 106(14), 5854–5858.
        # https://doi.org/10.1073/pnas.0901695106
        'permeability': 1e-6 * units.cm / units.sec,
        'surface_area_mass_ratio': 132 * units.cm**2 / units.mg,
        'volume_variable': 'volume',
    }

    def ports_schema(self):

        schema = {
            'internal': {
                # Molecule concentration in mmol/L
                molecule: {
                    '_default': self.parameters['default_default'],
                    '_divider': 'set',
                }
                for molecule in self.parameters['molecules_to_diffuse']
            },
            'external': {
                # Molecule concentration in mmol/L
                molecule: {
                    '_default': self.parameters['default_default'],
                    '_divider': 'set',
                }
                for molecule in self.parameters['molecules_to_diffuse']
            },
            'fields': {
                # Molecule count in mmol
                molecule: {
                    '_default': np.full(
                        (1, 1), self.parameters['default_default']),
                }
                for molecule in self.parameters['molecules_to_diffuse']
            },
            'global': {
                self.parameters['volume_variable']: {
                    '_default': self.parameters['default_default'],
                },
                'location': {
                    '_default': [0.5, 0.5],
                },
                'dry_mass': {
                    '_default': (
                        self.parameters['default_default'] * units.fg),
                },
            },
            'dimensions': {
                'bounds': {
                    '_default': [1, 1],
                },
                'n_bins': {
                    '_default': [1, 1],
                },
                'depth': {
                    '_default': 1,
                },
            },
        }

        for port, port_conf in self.parameters['default_state'].items():
            for variable, default in port_conf.items():
                if variable in schema[port]:
                    schema[port][variable]['_default'] = default

        return schema

    def next_update(self, timestep, states):
        permeability = self.parameters['permeability']
        area_mass = self.parameters['surface_area_mass_ratio']
        mass = states['global']['dry_mass']
        flux_mmol = {}
        for molecule in self.parameters['molecules_to_diffuse']:
            # Flux is positive when leaving the cell
            delta_concentration = (
                states['internal'][molecule]
                - states['external'][molecule]
            ) * units.mmol / units.L
            # Fick's first law of diffusion:
            rate = permeability * area_mass * delta_concentration
            flux = rate * mass * timestep * units.sec
            flux_mmol[molecule] = flux
        flux_counts = {
            molecule: flux * AVOGADRO
            for molecule, flux in flux_mmol.items()
        }
        volume_key = self.parameters['volume_variable']
        if not isinstance(states['global'][volume_key], Quantity):
            cell_volume = states['global'][volume_key] * units.fL
        else:
            cell_volume = states['global'][volume_key]
        update = {
            'fields': {
                molecule: {
                    '_value': mol_flux.to(units.count).magnitude,
                    '_updater': {
                        'updater': 'update_field_with_exchange',
                        'port_mapping': {
                            'global': 'global',
                            'dimensions': 'dimensions',
                        },
                    },
                }
                for molecule, mol_flux in flux_counts.items()
            },
            'internal': {
                molecule: - (
                    mol_flux / cell_volume
                ).to(units.mmol / units.L).magnitude
                for molecule, mol_flux in flux_mmol.items()
            },
        }
        return update



"""
Subclasses of DictWriter and DictReader that parse plaintext as JSON strings,
allowing for basic type parsing and fields that are dictionaries or lists.
"""

import os
import csv
import json
import re
import numpy as np

try:
    from future_builtins import filter
except ImportError:
    pass

from vivarium.library.units import units

TSV_DIALECT = csv.excel_tab

def array_to_list(value):
    if isinstance(value, np.ndarray):
        value = value.tolist()

    return value


class JsonWriter(csv.DictWriter):
    def __init__(self, *args, **kwargs):
        csv.DictWriter.__init__(
            self, quotechar = "'", quoting = csv.QUOTE_MINIMAL, lineterminator="\n", *args, **kwargs
            )

    def _dict_to_list(self, rowdict):
        return csv.DictWriter._dict_to_list(self, {
            key:json.dumps(array_to_list(value))
            for key, value in rowdict.viewitems()
            })


def split_units(field):
    try:
        attribute = re.search(r'(.*?) \(', field).group(1)
        units_value =  eval(re.search(r'\((.*?)\)', field).group(1))
        return (attribute, units_value)
    except AttributeError:
        return (field, None)


class JsonReader(csv.DictReader):
    def __init__(self, *args, **kwargs):
        csv.DictReader.__init__(
            self,
            quotechar = "\"",
            quoting = csv.QUOTE_MINIMAL,
            *args, **kwargs)

        # This is a hack to strip extra quotes from the field names
        # Not proud of it, but it works.
        self.fieldnames # called for side effect

        self._fieldnames = [
            fieldname.strip('"') for fieldname in self._fieldnames]

        self.field_mapping = {
            field: split_units(field)
            for field in self._fieldnames}


def load_tsv(path):
    with open(path, 'rU') as tsvfile:
        reader = JsonReader(
            filter(lambda x: x.lstrip()[0] != "#", tsvfile),  # Strip comments
            dialect=TSV_DIALECT)
        attr_list = []
        for row in reader:
            entry = {}
            for field in reader.fieldnames:
                fieldname, units = reader.field_mapping[field]
                value = row[field]

                try:
                    value = json.loads(value)
                except:
                    pass

                if not units is None:
                    value *= units

                entry[fieldname] = value
            attr_list.append(entry)
    return attr_list
